from __future__ import annotations

import fnmatch
import logging
from datetime import datetime
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd

from pyaerocom import const
from pyaerocom._lowlevel_helpers import merge_dicts
from pyaerocom.combine_vardata_ungridded import combine_vardata_ungridded
from pyaerocom.exceptions import (
    DataCoverageError,
    DataExtractionError,
    MetaDataError,
    StationCoordinateError,
    TimeMatchError,
    VarNotAvailableError,
)
from pyaerocom.geodesy import get_country_info_coords
from pyaerocom.helpers import merge_station_data, same_meta_dict, start_stop
from pyaerocom.metastandards import STANDARD_META_KEYS
from pyaerocom.stationdata import StationData
from pyaerocom.ungridded_data_metadata import UngriddedDataMetadata
from pyaerocom.units.datetime import TsType
from pyaerocom.units.units_helpers import get_unit_conversion_fac

logger = logging.getLogger(__name__)


class UngriddedData(UngriddedDataMetadata):
    """Class representing point-cloud data (ungridded)

    The data is organised in a 2-dimensional numpy array where the first index
    (rows) axis corresponds to individual measurements (i.e. one timestamp of
    one variable) and along the second dimension (containing 11 columns) the
    actual values are stored (in column 6) along with additional information,
    such as metadata index (can be used as key in :attr:`metadata` to access
    additional information related to this measurement), timestamp, latitude,
    longitude, altitude of instrument, variable index and, in case of 3D
    data (e.g. LIDAR profiles), also the altitude corresponding to the data
    value.

    Note
    ----

    That said, let's look at two examples.

    **Example 1**: Suppose you load 3 variables from 5 files, each of which
    contains 30 timestamps. This corresponds to a total of 3*5*30=450 data
    points and hence, the shape of the underlying numpy array will be 450x11.

    **Example 2**: 3 variables, 5 files, 30 timestamps, but each variable
    is height resolved, containing 100 altitudes => 3*5*30*100=4500 data points,
    thus, the final shape will be 4500x11.

    TODO
    ----
    Include unit attribute for each variable (in pyaerocom.io package: make
    sure to include units during ungridded read, if available)

    Attributes
    ----------
    metadata : dict[float, dict[str, Any]]
        dictionary containing meta information about the data. Keys are
        floating point numbers corresponding to each station, values are
        corresponding dictionaries containing station information.
    meta_idx : dict[float, dict[str, list[int]]]
        dictionary containing index mapping for each station and variable. Keys
        correspond to metadata key (float -> station, see :attr:`metadata`) and
        values are dictionaries containing keys specifying variable name and
        corresponding values are arrays or lists, specifying indices (rows) of
        these station / variable information in :attr:`_data`. Note: this
        information is redundant and is there to accelerate station data
        extraction since the data index matches for a given metadata block
        do not need to be searched in the underlying numpy array.
    var_idx : dict[str, float]
        mapping of variable name (keys, e.g. od550aer) to numerical variable
        index of this variable in data numpy array (in column specified by
        :attr:`_VARINDEX`)

    Parameters
    ----------
    num_points : :obj:`int`, optional
        initial number of total datapoints (number of rows in 2D dataarray)
    add_cols : :obj:`list`, optional
        list of additional index column names of 2D datarray.

    """

    #: version for caching, needs also updating when UngriddedDataMetadata has changed
    __version__ = "0.23"

    #: default number of rows that are dynamically added if total number of
    #: data rows is reached.
    _CHUNKSIZE = 10000000

    #: The following indices specify what the individual rows of the datarray
    #: are reserved for. These may be expanded when creating an instance of
    #: this class by providing a list of additional index names.
    _METADATAKEYINDEX = 0
    _TIMEINDEX = 1
    _LATINDEX = 2
    _LONINDEX = 3
    _ALTITUDEINDEX = 4  # altitude of measurement device
    _VARINDEX = 5
    _DATAINDEX = 6
    _DATAHEIGHTINDEX = 7
    _DATAERRINDEX = 8  # col where errors can be stored
    _DATAFLAGINDEX = 9  # can be used to store flags
    _STOPTIMEINDEX = 10  # can be used to store stop time of acq.
    _TRASHINDEX = 11  # index where invalid data can be moved to (e.g. when outliers are removed)

    # The following number denotes the kept precision after the decimal dot of
    # the location (e.g denotes lat = 300.12345)
    # used to code lat and long in a single number for a uniqueness test
    _LOCATION_PRECISION = 5
    _LAT_OFFSET = 90.0

    @property
    def _ROWNO(self):
        return self._data.shape[0]

    def __init__(self, num_points=None, add_cols=None):
        super().__init__()  # initialize metadata
        if num_points is None:
            num_points = self._CHUNKSIZE

        self._chunksize = num_points
        self._index = self._init_index(add_cols)

        # keep private, this is not supposed to be used by the user
        self._data = np.full([num_points, self._COLNO], np.nan)
        self.meta_idx = {}
        self._idx = -1

    @staticmethod
    def _from_raw_parts(
        data: npt.NDTArray[float],
        metadata: dict[float, dict[str, Any]],
        meta_idx: dict[float, dict[str, list[int]]],
        var_idx: dict[str, float],
    ) -> UngriddedData:
        data_obj = UngriddedData()

        data_obj._data = data
        data_obj.meta_idx = meta_idx
        data_obj.metadata = metadata
        data_obj.var_idx = var_idx

        return data_obj

    def _check_index(self):
        """Checks if all indices are assigned correctly"""
        assert len(self.meta_idx) == len(self.metadata), "Mismatch len(meta_idx) and len(metadata)"

        assert sum(self.meta_idx) == sum(
            self.metadata
        ), "Mismatch between keys of metadata dict and meta_idx dict"

        _varnums = self._data[:, self._VARINDEX]
        var_indices = np.unique(_varnums[~np.isnan(_varnums)])

        assert len(var_indices) == len(
            self.var_idx
        ), "Mismatch between number of variables in data array and var_idx attr."

        assert sum(var_indices) == sum(
            self.var_idx.values()
        ), "Mismatch between variable indices in data array and var_idx attr."

        vars_avail = self.var_idx

        for idx, meta in self.metadata.items():
            if "var_info" not in meta:
                if "variables" not in meta:
                    raise AttributeError(
                        f"Need either variables (list) or var_info (dict) "
                        f"in meta block {idx}: {meta}"
                    )
                meta["var_info"] = {}
                for v in meta["variables"]:
                    meta["var_info"][v] = {}

            var_idx = self.meta_idx[idx]
            for var, indices in var_idx.items():
                if len(indices) == 0:
                    continue  # no data assigned for this metadata index

                assert (
                    var in meta["var_info"]
                ), f"Var {var} is indexed in meta_idx[{idx}] but not in metadata[{idx}]"
                var_idx_data = np.unique(self._data[indices, self._VARINDEX])
                assert (
                    len(var_idx_data) == 1
                ), f"Found multiple variable indices for var {var}: {var_idx_data}"
                assert var_idx_data[0] == vars_avail[var], (
                    f"Mismatch between {var} index assigned in data and "
                    f"var_idx for {idx} in meta-block"
                )

    @staticmethod
    def from_station_data(
        stats: StationData, add_meta_keys: list[str] | None = None
    ) -> UngriddedData:
        """
        Create UngriddedData from input station data object(s)

        Parameters
        ----------
        stats : iterator or StationData
            input data object(s)
        add_meta_keys : list, optional
            list of metadata keys that are supposed to be imported from the
            input `StationData` objects, in addition to the default metadata
            retrieved via :func:`StationData.get_meta`.

        Raises
        ------
        ValueError
            if any of the input data objects is not an instance of
            :class:`StationData`.

        Returns
        -------
        UngriddedDataMeta
            ungridded data object created from input station data objects

        """
        if add_meta_keys is None:
            add_meta_keys = []
        elif isinstance(add_meta_keys, str):
            add_meta_keys = [add_meta_keys]
        elif not isinstance(add_meta_keys, list):
            raise ValueError(f"Invalid input for add_meta_keys {add_meta_keys}... need list")
        if isinstance(stats, StationData):
            stats = [stats]
        data_obj = UngriddedData()

        meta_key = 0.0
        idx = 0

        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        var_count_glob = -1
        for stat in stats:
            if isinstance(stat, dict):
                stat = StationData(**stat)
            elif not isinstance(stat, StationData):
                raise ValueError("Need instances of StationData or dicts")
            metadata[meta_key] = {}
            metadata[meta_key].update(
                stat.get_meta(force_single_value=False, quality_check=False, add_none_vals=True)
            )
            for key in add_meta_keys:
                try:
                    val = stat[key]
                except KeyError:
                    val = "undefined"

                metadata[meta_key][key] = val

            metadata[meta_key]["var_info"] = {}

            meta_idx[meta_key] = {}

            append_vars = list(stat.var_info)

            for var in append_vars:
                if var not in data_obj.var_idx:
                    var_count_glob += 1
                    var_idx = var_count_glob
                    data_obj.var_idx[var] = var_idx
                else:
                    var_idx = data_obj.var_idx[var]

                vardata = stat[var]

                if isinstance(vardata, pd.Series):
                    times = vardata.index
                    values = vardata.values
                else:
                    times = stat["dtime"]
                    values = vardata
                    if not len(times) == len(values):
                        raise ValueError

                times = np.asarray(
                    [
                        (
                            np.datetime64(x.replace(tzinfo=None), "s")
                            if isinstance(x, datetime)
                            else np.datetime64(x, "s")
                        )
                        for x in times
                    ]
                )
                times = times.astype(np.float64)

                num_times = len(times)
                # check if size of data object needs to be extended
                if (idx + num_times) >= data_obj._ROWNO:
                    # if totnum < data_obj._CHUNKSIZE, then the latter is used
                    data_obj.add_chunk(num_times)

                start = idx
                stop = start + num_times

                # write common meta info for this station (data lon, lat and
                # altitude are set to station locations)
                data_obj._data[start:stop, data_obj._LATINDEX] = stat["latitude"]
                data_obj._data[start:stop, data_obj._LONINDEX] = stat["longitude"]
                data_obj._data[start:stop, data_obj._ALTITUDEINDEX] = stat["altitude"]
                data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key

                # write data to data object
                data_obj._data[start:stop, data_obj._TIMEINDEX] = times

                data_obj._data[start:stop, data_obj._DATAINDEX] = values

                data_obj._data[start:stop, data_obj._VARINDEX] = var_idx

                if var in stat.data_flagged:
                    invalid = stat.data_flagged[var]
                    data_obj._data[start:stop, data_obj._DATAFLAGINDEX] = invalid

                if var in stat.data_err:
                    errs = stat.data_err[var]
                    data_obj._data[start:stop, data_obj._DATAERRINDEX] = errs

                var_info = stat["var_info"][var]
                metadata[meta_key]["var_info"][var] = {}
                metadata[meta_key]["var_info"][var].update(var_info)
                meta_idx[meta_key][var] = np.arange(start, stop)

                idx += num_times

            meta_key += 1

        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]

        data_obj._check_index()

        return data_obj

    def add_station_data(
        self, stat, meta_idx=None, data_idx=None, check_index=False
    ):  # pragma: no cover
        raise NotImplementedError("Coming at some point")
        if meta_idx is None:
            meta_idx = self.last_meta_idx + 1
        elif meta_idx in self.meta_idx:
            raise ValueError(
                f"Cannot add data at meta block index {meta_idx}, index already exists"
            )

        if data_idx is None:
            data_idx = self._data.shape[0]
        elif not np.all(np.isnan(self._data[data_idx, :])):
            raise ValueError(f"Cannot add data at data index {data_idx}, index already exists")

    @property
    def last_meta_idx(self):
        """
        Index of last metadata block
        """
        return np.max(list(self.meta_idx))

    @property
    def index(self):
        return self._index

    def _init_index(self, add_cols=None):
        """Init index mapping for columns in dataarray"""
        idx = dict(
            meta=self._METADATAKEYINDEX,
            time=self._TIMEINDEX,
            stoptime=self._STOPTIMEINDEX,
            latitude=self._LATINDEX,
            longitude=self._LONINDEX,
            altitude=self._ALTITUDEINDEX,
            varidx=self._VARINDEX,
            data=self._DATAINDEX,
            dataerr=self._DATAERRINDEX,
            dataaltitude=self._DATAHEIGHTINDEX,
            dataflag=self._DATAFLAGINDEX,
            trash=self._TRASHINDEX,
        )

        next_idx = max(idx.values()) + 1
        if add_cols is not None:
            if not isinstance(add_cols, list | tuple):
                raise ValueError("Invalid input for add_cols. Need list or tuple")
            for name in add_cols:
                if name in idx:
                    raise ValueError(
                        f"Cannot add new index with name {name} since "
                        f"this index already exists at column position {idx[name]}"
                    )
                idx[name] = next_idx
                next_idx += 1
        return idx

    @property
    def _COLNO(self):
        return len(self._index)

    @property
    def has_flag_data(self):
        """Boolean specifying whether this object contains flag data"""
        return (~np.isnan(self._data[:, self._DATAFLAGINDEX])).any()

    def copy(self):
        """Make a copy of this object

        Returns
        -------
        UngriddedData
            copy of this object

        Raises
        ------
        MemoryError
            if copy is too big to fit into memory together with existing
            instance
        """
        from copy import deepcopy

        new = UngriddedData()
        self._copy_metadata_to(new)
        new._data = np.copy(self._data)
        new.meta_idx = deepcopy(self.meta_idx)
        return new

    @property
    def shape(self):
        """Shape of data array"""
        return self._data.shape

    def add_chunk(self, size=None):
        """Extend the size of the data array

        Parameters
        ----------
        size : :obj:`int`, optional
            number of additional rows. If None (default) or smaller than
            minimum chunksize specified in attribute ``_CHUNKSIZE``, then the
            latter is used.
        """
        if size is None or size < self._chunksize:
            size = self._chunksize
        chunk = np.full([size, self._COLNO], np.nan)
        self._data = np.append(self._data, chunk, axis=0)
        logger.info(f"adding chunk, new array size ({self._data.shape})")

    def _get_stat_coords(self):
        meta_idx = []
        coords = []
        for idx, meta in self.metadata.items():
            try:
                lat, lon = meta["latitude"], meta["longitude"]
            except Exception:
                logger.warning(f"Could not retrieve lat lon coord at meta index {idx}")
                continue
            meta_idx.append(idx)
            coords.append((lat, lon))
        return (meta_idx, coords)

    def _check_set_country(self):
        """CHecks all metadata entries for availability of country information

        Deprecated - no longer used?

        Metadata blocks that are missing country entry will be updated based
        on country inferred from corresponding lat / lon coordinate. Uses
        :func:`pyaerocom.geodesy.get_country_info_coords` (library
        reverse-geocode) to retrieve countries. This may be errouneous
        close to country borders as it uses Euclidean distance based on a list
        of known locations.

        Note
        ----
        Metadata blocks that do not contain latitude and longitude entries are
        skipped.

        Returns
        -------
        list
            metadata entries where country was added
        list
            corresponding countries that were inferred from lat / lon
        """
        # HK 2025-04-30
        logger.warning(
            "This method is deprecated because no usage could be found, except for tests for check_country_available"
        )
        meta_idx, coords = self._get_stat_coords()
        info = get_country_info_coords(coords)
        meta_idx_updated = []
        countries = []

        for i, idx in enumerate(meta_idx):
            meta = self.metadata[idx]
            if "country" not in meta or meta["country"] is None:
                country = info[i]["country"]
                meta["country"] = country
                meta["country_code"] = info[i]["country_code"]
                meta_idx_updated.append(idx)
                countries.append(country)
        return (meta_idx_updated, countries)

    # TODO: see docstring
    def to_station_data(
        self,
        meta_idx,
        vars_to_convert=None,
        start=None,
        stop=None,
        freq=None,
        ts_type_preferred=None,
        merge_if_multi=True,
        merge_pref_attr=None,
        merge_sort_by_largest=True,
        insert_nans=False,
        allow_wildcards_station_name=True,
        add_meta_keys=None,
        resample_how=None,
        min_num_obs=None,
    ):
        """Convert data from one station to :class:`StationData`

        Todo
        ----
        - Review for retrieval of profile data (e.g. Lidar data)

        Parameters
        ----------
        meta_idx : float
            index of station or name of station.
        vars_to_convert : :obj:`list` or :obj:`str`, optional
            variables that are supposed to be converted. If None, use all
            variables that are available for this station
        start
            start time, optional (if not None, input must be convertible into
            pandas.Timestamp)
        stop
            stop time, optional (if not None, input must be convertible into
            pandas.Timestamp)
        freq : str
            pandas frequency string (e.g. 'D' for daily, 'M' for month end) or
            valid pyaerocom ts_type
        merge_if_multi : bool
            if True and if data request results in multiple instances of
            StationData objects, then these are attempted to be merged into one
            :class:`StationData` object using :func:`merge_station_data`
        merge_pref_attr
            only relevant for merging of multiple matches: preferred attribute
            that is used to sort the individual StationData objects by relevance.
            Needs to be available in each of the individual StationData objects.
            For details cf. :attr:`pref_attr` in docstring of
            :func:`merge_station_data`. Example could be `revision_date`. If
            None, then the stations will be sorted based on the number of
            available data points (if :attr:`merge_sort_by_largest` is True,
            which is default).
        merge_sort_by_largest : bool
            only relevant for merging of multiple matches: cf. prev. attr. and
            docstring of :func:`merge_station_data` method.
        insert_nans : bool
            if True, then the retrieved :class:`StationData` objects are filled
            with NaNs
        allow_wildcards_station_name : bool
            if True and if input `meta_idx` is a string (i.e. a station name or
            pattern), metadata matches will be identified applying wildcard
            matches between input `meta_idx` and all station names in this
            object.

        Returns
        -------
        StationData or list
            StationData object(s) containing results. list is only returned if
            input for meta_idx is station name and multiple matches are
            detected for that station (e.g. data from different instruments),
            else single instance of StationData. All variable time series are
            inserted as pandas Series
        """
        if isinstance(vars_to_convert, str):
            vars_to_convert = [vars_to_convert]
        elif vars_to_convert is None:
            vars_to_convert = self.contains_vars
            if len(vars_to_convert) == 0:
                raise DataCoverageError("UngriddedData object does not contain any variables")
        if start is None and stop is None:
            start = pd.Timestamp("1970")
            stop = pd.Timestamp("2200")
        else:
            start, stop = start_stop(start, stop)

        if isinstance(meta_idx, str):
            # user asks explicitly for station name, find all meta indices
            # that match this station
            meta_idx = self.find_station_meta_indices(meta_idx, allow_wildcards_station_name)
        if not isinstance(meta_idx, list):
            meta_idx = [meta_idx]

        stats = []
        # ToDo: check consistency, consider using methods in helpers.py
        # check also Hans' issue on the topic
        start, stop = np.datetime64(start), np.datetime64(stop)

        for idx in meta_idx:
            try:
                stat = self._metablock_to_stationdata(
                    idx, vars_to_convert, start, stop, add_meta_keys
                )
                if ts_type_preferred is not None:
                    if "ts_type" in stat["var_info"][vars_to_convert[0]].keys():
                        if TsType(stat["var_info"][vars_to_convert[0]]["ts_type"]) < TsType(
                            ts_type_preferred
                        ):
                            continue
                    elif "ts_type" in stat.keys():
                        if TsType(stat["ts_type"]) < TsType(ts_type_preferred):
                            continue
                    else:
                        raise KeyError("Could not find ts_type in stat")
                stats.append(stat)
            except (VarNotAvailableError, DataCoverageError) as e:
                logger.debug(f"Skipping meta index {idx}. Reason: {repr(e)}")
        if merge_if_multi and len(stats) > 1:
            if len(vars_to_convert) > 1:
                raise NotImplementedError(
                    "Cannot yet merge multiple stations with multiple variables."
                )
            if merge_pref_attr is None:
                merge_pref_attr = self._try_infer_stat_merge_pref_attr(stats)
            merged = merge_station_data(
                stats,
                vars_to_convert,
                pref_attr=merge_pref_attr,
                sort_by_largest=merge_sort_by_largest,
                fill_missing_nan=False,
                resample_how=resample_how,
                min_num_obs=min_num_obs,
            )
            stats = [merged]

        stats_ok = []
        for stat in stats:
            for var in vars_to_convert:
                if var not in stat:
                    continue
                if freq is not None:
                    stat.resample_time(
                        var,
                        freq,
                        how=resample_how,
                        min_num_obs=min_num_obs,
                        inplace=True,
                    )
                elif insert_nans:
                    stat.insert_nans_timeseries(var)
                if np.all(np.isnan(stat[var].values)):
                    stat = stat.remove_variable(var)
            if any([x in stat for x in vars_to_convert]):
                stats_ok.append(stat)

        if len(stats_ok) == 0:
            raise DataCoverageError(
                f"{vars_to_convert} data could not be retrieved "
                f"for meta index (or station name) {meta_idx}"
            )
        elif len(stats_ok) == 1:
            # return StationData object and not list
            return stats_ok[0]
        return stats_ok

    ### TODO: check if both `variables` and `var_info` attrs are required in
    ### metdatda blocks
    def _metablock_to_stationdata(
        self, meta_idx, vars_to_convert, start=None, stop=None, add_meta_keys=None
    ):
        """Convert one metadata index to StationData (helper method)

        See :func:`to_station_data` for input parameters
        """
        if add_meta_keys is None:
            add_meta_keys = []
        elif isinstance(add_meta_keys, str):
            add_meta_keys = [add_meta_keys]

        sd = StationData()
        meta = self.metadata[meta_idx]

        # TODO: make sure in reading classes that data_revision is assigned
        # to each metadata block and not only in self.data_revision
        rev = None
        if "data_revision" in meta:
            rev = meta["data_revision"]
        else:
            try:
                rev = self.get_data_revision[meta["data_id"]]
            except Exception:
                logger.debug("Data revision could not be accessed")
        sd.data_revision = rev
        try:
            vars_avail = list(meta["var_info"])
        except KeyError:
            if "variables" not in meta or meta["variables"] in (None, []):
                raise VarNotAvailableError("Metablock does not contain variable information")
            vars_avail = meta["variables"]

        for key in STANDARD_META_KEYS + add_meta_keys:
            if key in sd.PROTECTED_KEYS:
                logger.warning(f"skipping protected key: {key}")
                continue
            try:
                sd[key] = meta[key]
            except KeyError:
                pass

        try:
            sd["ts_type_src"] = meta["ts_type"]
        except KeyError:
            pass

        # assign station coordinates explicitly
        for ck in sd.STANDARD_COORD_KEYS:
            try:
                sd.station_coords[ck] = meta[ck]
            except KeyError:
                pass
        # if no input variables are provided, use the ones that are available
        # for this metadata block
        if vars_to_convert is None:
            vars_to_convert = vars_avail

        # find overlapping variables (ignore all other ones)
        vars_avail = np.intersect1d(vars_to_convert, vars_avail)
        if not len(vars_avail) >= 1:
            raise VarNotAvailableError(
                "None of the input variables matches, or station does not contain data."
            )
        # init helper boolean that is set to True if valid data can be found
        # for at least one of the input variables
        FOUND_ONE = False
        for var in vars_avail:
            # get indices of this variable
            var_idx = self.meta_idx[meta_idx][var]

            # vector of timestamps corresponding to this variable
            dtime = self._data[var_idx, self._TIMEINDEX].astype("datetime64[s]")

            # get subset
            subset = self._data[var_idx]

            # make sure to extract only valid timestamps
            if start is None:
                start = dtime.min()
            if stop is None:
                stop = dtime.max()

            # create access mask for valid time stamps
            tmask = np.logical_and(dtime >= start, dtime <= stop)

            # make sure there is some valid data
            if tmask.sum() == 0:
                logger.debug(
                    f"Ignoring station {sd['station_name']}, var {var} ({sd['data_id']}): "
                    f"no data available in specified time interval {start} - {stop}"
                )
                continue

            dtime = dtime[tmask]
            subset = subset[tmask]

            vals = subset[:, self._DATAINDEX]
            if np.all(np.isnan(vals)):
                logger.debug(
                    f"Ignoring station {sd['station_name']}, var {var} ({sd['data_id']}): "
                    f"All values are NaN"
                )
                continue
            vals_err = subset[:, self._DATAERRINDEX]
            flagged = subset[:, self._DATAFLAGINDEX]
            altitude = subset[:, self._DATAHEIGHTINDEX]

            data = pd.Series(vals, dtime)
            if not data.index.is_monotonic_increasing:
                data = data.sort_index()
            if any(~np.isnan(vals_err)):
                sd.data_err[var] = vals_err
            if any(~np.isnan(flagged)):
                sd.data_flagged[var] = flagged

            sd["dtime"] = data.index.values
            sd[var] = data
            sd["var_info"][var] = {}
            FOUND_ONE = True
            # check if there is information about altitude (then relevant 3D
            # variables and parameters are included too)
            if "var_info" in meta:
                vi = meta["var_info"]
            else:
                vi = {}
            if not np.isnan(altitude).all():
                if "altitude" in vi:
                    sd.var_info["altitude"] = vi["altitude"]
                sd.altitude = altitude
            if var in vi:
                sd.var_info[var].update(vi[var])

            if len(data.index) == len(data.index.unique()):
                sd.var_info[var]["overlap"] = False
            else:
                sd.var_info[var]["overlap"] = True
        if not FOUND_ONE:
            raise DataCoverageError(
                f"Could not retrieve any valid data for station {sd['station_name']} "
                f"and input variables {vars_to_convert}"
            )
        return sd

    def _generate_station_index(self, by_station_name=True, ignore_index=None):
        """Generates index to loop over station names or metadata block indices"""
        if ignore_index is None:
            if by_station_name:
                return self.unique_station_names  # all station names
            return list(range(len(self.metadata)))  # all meta indices

        if not by_station_name:
            from pyaerocom.helpers import isnumeric

            if isnumeric(ignore_index):
                ignore_index = [ignore_index]
            if not isinstance(ignore_index, list):
                raise ValueError("Invalid input for ignore_index, need number or list")
            return [i for i in range(len(self.metadata)) if i not in ignore_index]

        # by station name and ignore certation stations
        _iter = []
        if isinstance(ignore_index, str):
            ignore_index = [ignore_index]
        if not isinstance(ignore_index, list):
            raise ValueError("Invalid input for ignore_index, need str or list")
        for stat_name in self.unique_station_names:
            ok = True
            for name_or_pattern in ignore_index:
                if fnmatch.fnmatch(stat_name, name_or_pattern):
                    ok = False
            if ok:
                _iter.append(stat_name)
        return _iter

    def to_station_data_all(
        self,
        vars_to_convert=None,
        start=None,
        stop=None,
        freq=None,
        ts_type_preferred=None,
        by_station_name=True,
        ignore_index=None,
        **kwargs,
    ):
        """Convert all data to :class:`StationData` objects

        Creates one instance of :class:`StationData` for each metadata block in
        this object.

        Parameters
        ----------
        vars_to_convert : :obj:`list` or :obj:`str`, optional
            variables that are supposed to be converted. If None, use all
            variables that are available for this station
        start
            start time, optional (if not None, input must be convertible into
            pandas.Timestamp)
        stop
            stop time, optional (if not None, input must be convertible into
            pandas.Timestamp)
        freq : str
            pandas frequency string (e.g. 'D' for daily, 'M' for month end)
            or valid pyaerocom ts_type (e.g. 'hourly', 'monthly').
        by_station_name : bool
            if True, then iter over unique_station_name (and merge multiple
            matches if applicable), else, iter over metadata index
        **kwargs
            additional keyword args passed to :func:`to_station_data` (e.g.
            `merge_if_multi, merge_pref_attr, merge_sort_by_largest,
            insert_nans`)

        Returns
        -------
        dict
            5-element dictionary containing following key / value pairs:

                - stats: list of :class:`StationData` objects
                - station_name: list of corresponding station names
                - station_type: list of corresponding station types, might be empty
                - latitude: list of latitude coordinates
                - longitude: list of longitude coordinates

        """
        out_data = {
            "stats": [],
            "station_name": [],
            "station_type": [],
            "latitude": [],
            "failed": [],
            "longitude": [],
        }

        _iter = self._generate_station_index(by_station_name, ignore_index)
        for idx in _iter:
            try:
                data = self.to_station_data(
                    idx,
                    vars_to_convert,
                    start,
                    stop,
                    freq,
                    merge_if_multi=True,
                    allow_wildcards_station_name=False,
                    ts_type_preferred=ts_type_preferred,
                    **kwargs,
                )
                out_data["latitude"].append(data["latitude"])
                out_data["longitude"].append(data["longitude"])
                out_data["station_name"].append(data["station_name"])
                if hasattr(data, "station_type"):
                    out_data["station_type"].append(data["station_type"])
                else:
                    logger.debug(
                        "No station_type found in StationData, station_type will be blank"
                    )
                out_data["stats"].append(data)

            # catch the exceptions that are acceptable
            except (
                VarNotAvailableError,
                TimeMatchError,
                DataCoverageError,
                NotImplementedError,
                StationCoordinateError,
            ) as e:
                logger.debug(f"Failed to convert to StationData Error: {repr(e)}")
                out_data["failed"].append([idx, repr(e)])
        return out_data

    def check_convert_var_units(self, var_name, to_unit=None, inplace=True):
        obj = self if inplace else self.copy()

        # get the unit
        if to_unit is None:
            to_unit = const.VARS[var_name]["units"]

        for i, meta in obj.metadata.items():
            if var_name in meta["var_info"]:
                try:
                    unit = meta["var_info"][var_name]["units"]
                except KeyError:
                    add_str = ""
                    if "unit" in meta["var_info"][var_name]:
                        add_str = (
                            "Corresponding var_info dict contains "
                            'attr. "unit", which is deprecated, please '
                            "check corresponding reading routine. "
                        )
                    raise MetaDataError(
                        f"Failed to access unit information for variable {var_name} "
                        f"in metadata block {i}. {add_str}"
                    )
                fac = get_unit_conversion_fac(unit, to_unit, var_name)
                if fac != 1:
                    meta_idx = obj.meta_idx[i][var_name]
                    current = obj._data[meta_idx, obj._DATAINDEX]
                    new = current * fac
                    obj._data[meta_idx, obj._DATAINDEX] = new
                    obj.metadata[i]["var_info"][var_name]["units"] = to_unit

        return obj

    def set_flags_nan(self, inplace=False):
        """Set all flagged datapoints to NaN

        Parameters
        ----------
        inplace : bool
            if True, the flagged datapoints will be set to NaN in this object,
            otherwise a new object will be created and returned

        Returns
        -------
        UngriddedData
            data object that has all flagged data values set to NaN

        Raises
        ------
        AttributeError
            if no flags are assigned
        """

        if not self.has_flag_data:
            raise AttributeError("Ungridded data object does not contain flagged data points")
        if inplace:
            obj = self
        else:
            obj = self.copy()
        mask = obj._data[:, obj._DATAFLAGINDEX] == 1

        obj._data[mask, obj._DATAINDEX] = np.nan
        obj._add_to_filter_history("set_flags_nan")
        return obj

    def remove_outliers(
        self,
        var_name,
        inplace=False,
        low=None,
        high=None,
        unit_ref=None,
        move_to_trash=True,
    ):
        """Method that can be used to remove outliers from data

        Parameters
        ----------
        var_name : str
            variable name
        inplace : bool
            if True, the outliers will be removed in this object, otherwise
            a new object will be created and returned
        low : float
            lower end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. minimum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        high : float
            upper end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. maximum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        unit_ref : str
            reference unit for assessment of input outlier ranges: all data
            needs to be in that unit, else an Exception will be raised
        move_to_trash : bool
            if True, then all detected outliers will be moved to the trash
            column of this data object (i.e. column no. specified at
            :attr:`UngriddedData._TRASHINDEX`).

        Returns
        -------
        UngriddedData
            ungridded data object that has all outliers for this variable
            removed.

        Raises
        ------
        ValueError
            if input :attr:`move_to_trash` is True and in case for some of the
            measurements there is already data in the trash.
        """
        if inplace:
            new = self
        else:
            new = self.copy()

        new.check_convert_var_units(var_name, to_unit=unit_ref)

        if low is None:
            low = const.VARS[var_name].minimum
            logger.info(f"Setting {var_name} outlier lower lim: {low:.2f}")
        if high is None:
            high = const.VARS[var_name].maximum
            logger.info(f"Setting {var_name} outlier upper lim: {high:.2f}")
        var_idx = new.var_idx[var_name]
        var_mask = new._data[:, new._VARINDEX] == var_idx

        all_data = new._data[:, new._DATAINDEX]
        invalid_mask = np.logical_or(all_data < low, all_data > high)

        mask = invalid_mask * var_mask
        invalid_vals = new._data[mask, new._DATAINDEX]
        new._data[mask, new._DATAINDEX] = np.nan

        if move_to_trash:
            # check if trash is empty and put outliers into trash
            trash = new._data[mask, new._TRASHINDEX]
            if np.isnan(trash).sum() == len(trash):  # trash is empty
                new._data[mask, new._TRASHINDEX] = invalid_vals
            else:
                raise ValueError(
                    "Trash is not empty for some of the datapoints. "
                    "Please empty trash first using method "
                    ":func:`empty_trash` or deactivate input arg "
                    ":attr:`move_to_trash`"
                )

        new._add_to_filter_history(
            f"Removed {len(invalid_vals)} outliers from {var_name} data "
            f"(range: {low}-{high}, in trash: {move_to_trash})"
        )
        return new

    def empty_trash(self):
        """Set all values in trash column to NaN"""
        self._data[:, self._TRASHINDEX] = np.nan

    def _new_from_meta_blocks(self, meta_indices, totnum_new):
        # make a new empty object with the right size (totnum_new)

        new = UngriddedData(num_points=totnum_new)

        meta_idx_new = 0.0
        data_idx_new = 0

        # loop over old meta_idx and extract data and create new meta_idx in
        # output data object
        for meta_idx in meta_indices:
            meta = self.metadata[meta_idx]
            new.metadata[meta_idx_new] = meta
            new.meta_idx[meta_idx_new] = {}
            for var in meta["var_info"]:
                if var in self.ALLOWED_VERT_COORD_TYPES:
                    continue
                indices = self.meta_idx[meta_idx][var]
                totnum = len(indices)

                stop = data_idx_new + totnum
                while stop > new._data.shape[0]:
                    new.add_chunk()
                new._data[data_idx_new:stop, :] = self._data[indices, :]
                new._data[data_idx_new:stop, new._METADATAKEYINDEX] = meta_idx_new
                new.meta_idx[meta_idx_new][var] = np.arange(data_idx_new, stop)
                new.var_idx[var] = self.var_idx[var]
                data_idx_new += totnum

            meta_idx_new += 1

        if meta_idx_new == 0 or data_idx_new == 0:
            raise DataExtractionError("Filtering results in empty data object")
        new._data = new._data[:data_idx_new]

        # write history of filtering applied
        new.filter_hist.update(self.filter_hist)
        new._data_revision.update(self._data_revision)

        return new

    def _len_datapoints(self, meta_idx, var):
        """Get the number of datapoints for meta_idx and var."""
        if isinstance(meta_idx, float):
            meta_idx = [meta_idx]
        if isinstance(var, str):
            var = [var]
        totnum = 0
        for m in meta_idx:
            for v in var:
                try:
                    totnum += len(self.meta_idx[m][v])
                except KeyError:
                    logger.debug(
                        f"Ignoring variable {var} in meta block {meta_idx} "
                        f"since no data could be found"
                    )
        return

    def clear_meta_no_data(self, inplace=True):
        """Remove all metadata blocks that do not have data associated with it

        Parameters
        ----------
        inplace : bool
            if True, the changes are applied to this instance directly, else
            to a copy

        Returns
        -------
        UngriddedData
            cleaned up data object

        Raises
        ------
        DataCoverageError
            if filtering results in empty data object
        """
        if inplace:
            obj = self
        else:
            obj = self.copy()
        meta_new = {}
        meta_idx_new = {}
        for idx, val in obj.meta_idx.items():
            meta = obj.metadata[idx]
            if not bool(val):  # no data assigned with this metadata block
                # sanity check
                if bool(meta["var_info"]):
                    raise AttributeError(
                        "meta_idx {} suggests empty data block "
                        "but metadata[{}] contains variable "
                        "information"
                    )
            else:
                meta_new[idx] = meta
                meta_idx_new[idx] = val
        num_removed = len(obj.metadata) - len(meta_new)
        if not bool(meta_new):
            raise DataCoverageError("UngriddedData object appears to be empty")
        elif num_removed > 0:  # some meta blocks are empty
            obj.metadata = meta_new
            obj.meta_idx = meta_idx_new

        obj._add_to_filter_history(
            f"Removed {num_removed} metadata blocks that have no data assigned"
        )
        obj._check_index()
        return obj

    def extract_var(self, var_name, check_index=True):
        """Split this object into single-var UngriddedData objects

        Parameters
        ----------
        var_name : str
            name of variable that is supposed to be extracted
        check_index : Bool
            Call :func:`_check_index` in the new data object.

        Returns
        -------
        UngriddedData
            new data object containing only input variable data
        """
        if var_name not in self.contains_vars:
            # try alias
            _var = const.VARS[var_name].var_name_aerocom
            if _var in self.contains_vars:
                var_name = _var
            else:
                raise VarNotAvailableError(f"No such variable {var_name} in data")
        elif len(self.contains_vars) == 1:
            logger.info("Data object is already single variable. Returning copy")
            return self.copy()

        var_idx = self.var_idx[var_name]

        totnum = np.sum(self._data[:, self._VARINDEX] == var_idx)

        colnum, rownum = self.shape

        if rownum != len(self._init_index()):
            raise NotImplementedError(
                "Cannot split UngriddedData objects that have "
                "additional columns other than default columns"
            )

        subset = UngriddedData(totnum)

        subset.var_idx[var_name] = 0
        subset._index = self.index

        meta_idx = -1
        arr_idx = 0

        for midx, didx in self.meta_idx.items():
            if var_name in didx and len(didx[var_name]) > 0:
                meta_idx += 1
                meta = {}
                _meta = self.metadata[midx]
                meta.update(_meta)
                meta["var_info"] = {}
                meta["var_info"][var_name] = _meta["var_info"][var_name]
                meta["variables"] = [var_name]
                subset.metadata[meta_idx] = meta

                idx = didx[var_name]

                subset.meta_idx[meta_idx] = {}

                num_add = len(idx)
                start = arr_idx
                stop = arr_idx + num_add
                subset.meta_idx[meta_idx][var_name] = np.arange(start, stop)

                subset._data[start:stop] = self._data[idx]
                subset._data[start:stop, subset._METADATAKEYINDEX] = meta_idx
                subset._data[start:stop, subset._VARINDEX] = 0

                arr_idx += num_add

        if check_index:
            subset._check_index()
        subset.filter_hist.update(self.filter_hist)
        subset._add_to_filter_history(
            f"Created {var_name} single var object from multivar UngriddedData instance"
        )
        return subset

    def extract_vars(self, var_names, check_index=True):
        """Extract multiple variables from dataset

        Loops over input variable names and calls :func:`extract_var` to
        retrieve single variable UngriddedData objects for each variable and
        then merges all of these into one object

        Parameters
        ----------
        var_names : list or str
            list of variables to be extracted
        check_index : Bool
            Call :func:`_check_index` in the new data object.

        Returns
        -------
        UngriddedData
            new data object containing input variables

        Raises
        -------
        VarNotAvailableError
            if one of the input variables is not available in this data
            object
        """
        if isinstance(var_names, str):
            return self.extract_var(var_names)
        data = UngriddedData()

        for var in var_names:
            data.append(self.extract_var(var, check_index=False))
        if check_index:
            data._check_index()
        return data

    def _find_common_meta(self, ignore_keys=None):
        """Searches all metadata dictionaries that are the same

        Parameters
        ----------
        ignore_keys : list
            list containing meta keys that are supposed to be ignored

        Returns
        -------
        tuple
            2-element tuple containing

            - list containing lists with common meta indices
            - list containing corresponding meta dictionaries
        """
        if ignore_keys is None:
            ignore_keys = []
        meta_registered = []
        same_indices = []
        for meta_key, meta in self.metadata.items():
            found = False
            for idx, meta_reg in enumerate(meta_registered):
                if same_meta_dict(meta_reg, meta, ignore_keys=ignore_keys):
                    same_indices[idx].append(meta_key)
                    found = True

            if not found:
                meta_registered.append(meta)
                same_indices.append([meta_key])

        return same_indices

    def merge_common_meta(self, ignore_keys=None):
        """Merge all meta entries that are the same

        Note
        ----
        If there is an overlap in time between the data, the blocks are not
        merged

        Todo
        ----
        Keep mapping of ``var_info`` (if defined in ``metadata``) to data
        points (e.g. EBAS), since the data sources may be at different
        wavelengths.

        Parameters
        ----------
        ignore_keys : list
            list containing meta keys that are supposed to be ignored

        Returns
        -------
        UngriddedData
            merged data object
        """
        if ignore_keys is None:
            ignore_keys = []
        sh = self.shape
        lst_meta_idx = self._find_common_meta(ignore_keys)
        new = UngriddedData(num_points=self.shape[0])
        didx = 0
        for i, idx_lst in enumerate(lst_meta_idx):
            _meta_check = {}
            # write metadata of first index that matches
            _meta_check.update(self.metadata[idx_lst[0]])
            _meta_idx_new = {}
            for j, meta_idx in enumerate(idx_lst):
                if j > 0:  # don't check first against first
                    meta = self.metadata[meta_idx]
                    merged = merge_dicts(meta, _meta_check)
                    for key in ignore_keys:
                        _meta_check[key] = merged[key]

                data_var_idx = self.meta_idx[meta_idx]
                for var, data_idx in data_var_idx.items():
                    num = len(data_idx)
                    stop = didx + num
                    new._data[didx:stop, :] = self._data[data_idx]
                    new._data[didx:stop, 0] = i
                    if var not in _meta_idx_new:
                        _meta_idx_new[var] = np.arange(didx, stop)
                    else:
                        _idx = np.append(_meta_idx_new[var], np.arange(didx, stop))
                        _meta_idx_new[var] = _idx
                    didx += num

            new.meta_idx[i] = _meta_idx_new
            new.metadata[i] = _meta_check
        new.var_idx.update(self.var_idx)
        new.filter_hist.update(self.filter_hist)
        if not new.shape == sh:
            raise Exception(
                "FATAL: Mismatch in shape between initial and "
                "and final object. Developers: please check"
            )
        return new

    def append_station_data(self, stats, add_meta_keys=None):
        raise NotImplementedError()

    def merge(self, other, new_obj=True):
        """Merge another data object with this one

        Parameters
        -----------
        other : UngriddedData
            other data object
        new_obj : bool
            if True, this object remains unchanged and the merged data objects
            are returned in a new instance of :class:`UngriddedData`. If False,
            then this object is modified

        Returns
        -------
        UngriddedData
            merged data object

        Raises
        -------
        ValueError
            if input object is not an instance of :class:`UngriddedData`
        """
        if not isinstance(other, UngriddedData):
            raise ValueError(
                f"Cannot add any UngriddedDataContainer to UngriddedData, got: {type(other)}, please switch order, or implement UngriddedData.append_station_data"
            )
        if new_obj:
            obj = self.copy()
        else:
            obj = self

        if obj.is_empty:
            obj._data = other._data
            obj.metadata = other.metadata
            # obj.unit = other.unit
            obj._data_revision = other._data_revision
            obj.meta_idx = other.meta_idx
            # potentially temporary fix for pyaro actrisebas reader
            # if len(other.meta_idx) != len(other.metadata):
            #     obj.meta_idx = {key: other.meta_idx[key] for key in other.metadata.keys()}
            obj.var_idx = other.var_idx
        else:
            # get offset in metadata index
            meta_offset = max(obj.metadata) + 1
            data_offset = obj.shape[0]

            # add this offset to indices of meta dictionary in input data object
            for meta_idx_other, meta_other in other.metadata.items():
                meta_idx = meta_offset + meta_idx_other
                obj.metadata[meta_idx] = meta_other
                _idx_map = {}
                for var_name, indices in other.meta_idx[meta_idx_other].items():
                    _idx_map[var_name] = np.asarray(indices) + data_offset
                obj.meta_idx[meta_idx] = _idx_map

            for var, idx in other.var_idx.items():
                if var in obj.var_idx:  # variable already exists in this object
                    if not idx == obj.var_idx[var]:
                        other._change_var_idx(var, obj.var_idx[var])
                else:  # variable does not yet exist
                    idx_exists = [v for v in obj.var_idx.values()]
                    if idx in idx_exists:
                        # variable index is already assigned to another
                        # variable and needs to be changed
                        new_idx = max(idx_exists) + 1
                        other._change_var_idx(var, new_idx)
                        obj.var_idx[var] = new_idx
                    else:
                        obj.var_idx[var] = idx
            obj._data = np.vstack([obj._data, other._data])
            obj._data_revision.update(other._data_revision)
        obj.filter_hist.update(other.filter_hist)
        obj._check_index()
        return obj

    def colocate_vardata(
        self, var1, data_id1=None, var2=None, data_id2=None, other=None, **kwargs
    ):
        # UNTESTED UNDOCUMENTED METHOD, HK 2025-03-15
        if other is None:
            other = self
        if var2 is None:
            var2 = var1
        if data_id1 is None:
            contains = self.contains_datasets
            if len(contains) > 1:
                raise ValueError(
                    "Please provide data_id1 since data object contains more than 1 dataset..."
                )
            data_id1 = contains[0]

        if data_id2 is None:
            contains = other.contains_datasets
            if len(contains) > 1:
                raise ValueError(
                    "Please provide data_id2 since data object contains more than 1 dataset..."
                )
            data_id2 = contains[0]
        if self is other and data_id1 == data_id2 and var1 == var2:
            raise ValueError(
                "Input combination too unspecific, please provide "
                "either another data object, 2 different data IDs "
                "or 2 different variable names"
            )
        input_data = [(self, data_id1, var1), (other, data_id2, var2)]
        statlist = combine_vardata_ungridded(input_data, **kwargs)

        new = UngriddedData.from_station_data(statlist)
        return new

    def _change_var_idx(self, var_name, new_idx):
        """Change index that is assigned to variable

        Each variable in this object has assigned a unique index that is
        stored in the dictionary :attr:`var_idx` and which is used internally
        to access data from a certain variable from the data array
        :attr:`_data` (the indices are stored in the data column specified by
        :attr:`_VARINDEX`, cf. class header).

        This index thus needs to be unique for each variable and hence, may
        need to be updated, when two instances of :class:`UngriddedData` are
        merged (cf. :func:`merge`).

        And the latter is exactrly what this function does.

        Parameters
        ----------
        var_name : str
            name of variable
        new_idx : int
            new index of variable

        Raises
        ------
        ValueError
            if input ``new_idx`` already exist in this object as a variable
            index
        """
        if new_idx in self.var_idx.values():
            raise ValueError(
                "Fatal: variable index cannot be assigned a new "
                "index that is already assigned to one of the "
                "variables in this object"
            )
        cidx = self.var_idx[var_name]
        self.var_idx[var_name] = new_idx
        var_indices = np.where(self._data[:, self._VARINDEX] == cidx)
        self._data[var_indices, self._VARINDEX] = new_idx

    def all_datapoints_var(self, var_name):
        """Get array of all data values of input variable

        Parameters
        ----------
        var_name : str
            variable name

        Returns
        -------
        ndarray
            1-d numpy array containing all values of this variable

        Raises
        ------
        AttributeError
            if variable name is not available
        """
        if var_name not in self.var_idx:
            raise AttributeError(f"Variable {var_name} not available in data")
        idx = self.var_idx[var_name]
        mask = np.where(self._data[:, self._VARINDEX] == idx)[0]
        return self._data[mask, self._DATAINDEX]

    def __iter__(self):
        return self

    #: ToDo revise cases of DataCoverageError
    def __next__(self):
        self._idx += 1
        if self._idx == len(self.metadata):
            self._idx = -1
            raise StopIteration
        try:
            return self[self._idx]
        except DataCoverageError:
            logger.debug(
                f"No variable data in metadata block {self._idx}. Returning empty StationData"
            )
            return StationData()

    def __repr__(self):
        return f"{type(self).__name__} <networks: {self.contains_datasets}; vars: {self.contains_vars}; instruments: {self.contains_instruments}; No. of metadata units: {len(self.metadata)}"


def reduce_array_closest(arr_nominal, arr_to_be_reduced):
    logger.warning("This method is deprecated because no usage could be found, HK 2025-03-13")

    test = sorted(arr_to_be_reduced)
    closest_idx = []
    for num in sorted(arr_nominal):
        idx = np.argmin(abs(test - num))
        closest_idx.append(idx)
        test = test[(idx + 1) :]
    return closest_idx



import xarray
from pytest import approx
import numpy as np

from pyaerocom.projection_information import ProjectionInformation
from tests.fixtures.data_access import TEST_DATA

ROOT = TEST_DATA["MODELS"].path


def test_projection_information():
    with xarray.open_dataset(
        str(ROOT / "emep4no20240630" / "Base_hour.nc"), decode_timedelta=True
    ) as ds:
        pi = ProjectionInformation.from_xarray(ds, "SURF_ug_PM10_rh50")
        assert pi is not None
        assert pi.x_axis == "i"
        assert pi.y_axis == "j"
        x0, y0 = float(ds["i"][0]), float(ds["j"][0])
        lon0, lat0 = float(ds["lon"][0, 0]), float(ds["lat"][0, 0])
        (lat, lon) = pi.to_latlon(x0, y0)
        assert lat == approx(lat0, abs=1e-3)
        assert lon == approx(lon0, abs=1e-3)

        x, y = pi.to_proj(lat, lon)
        assert x == approx(x0, abs=1.0)
        assert y == approx(y0, abs=1.0)


def test_projection_snap():
    x = [10, 12, 13]
    y = [60, 61]

    lons, lats = np.meshgrid(x, y)

    field = np.ones_like(lons)

    ds = xarray.Dataset(
        {
            "x": xarray.DataArray(x, coords={"x": x}, attrs={"units": "degrees_east"}),
            "y": xarray.DataArray(y, coords={"y": y}, attrs={"units": "degrees_north"}),
            "lons": xarray.DataArray(lons, dims=["y", "x"], attrs={"units": "degrees_east"}),
            "lats": xarray.DataArray(lats, dims=["y", "x"], attrs={"units": "degrees_north"}),
            "field": xarray.DataArray(
                field,
                dims=["y", "x"],
                attrs={"grid_mapping": "projection", "coordinates": "longitude latitude"},
            ),
            "projection": xarray.DataArray(0, attrs={"grid_mapping_name": "latitude_longitude"}),
        }
    )
    pi = ProjectionInformation.from_xarray(ds, "field")
    assert pi._x_axis == "x"
    assert pi._y_axis == "y"

    # Transform like an identity matrix
    y0, x0 = 11, 61.5
    (lat, lon) = pi.to_latlon(x0, y0)
    assert x0 == approx(lon, abs=1e-3)
    assert y0 == approx(lat, abs=1e-3)



import logging
from pyaerocom.units.units import UnitConversionCallbackInfo

logger = logging.getLogger(__name__)


class LoggingCallback:
    """Class intended to be used as a callback for unit conversion
    in pyaerocom.units.Unit. It logs as either DEBUG or LOG depending
    on if the values where changed (ie. if factor is 1).
    """

    def __init__(self, log: logging.Logger | None = None) -> None:
        """
        :param log: Instance of logger to be used. If not provided, the logger
            of this module will be used, which may be undesired due to hiding the
            real source of the error so it is recommended to provide your own.
        """
        if log is None:
            log = logger
        self._logger = log

    def __call__(self, info: UnitConversionCallbackInfo) -> None:
        if info.factor == 1:
            level = logging.DEBUG
        else:
            level = logging.INFO

        self._logger.log(
            level,
            "Successfully converted unit of variable '%s' from '%s' to '%s' using conversion factor '%.2f'.",
            info.from_aerocom_var,
            info.from_cf_unit,
            info.to_cf_unit,
            info.factor,
        )



from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
)
from typing import Literal


class WebDisplaySetup(BaseModel):
    # Pydantic ConfigDict
    model_config = ConfigDict(protected_namespaces=())
    # WebDisplaySetup attributes
    map_zoom: Literal["World", "Europe", "xEMEP"] = "World"
    regions_how: Literal["default", "aerocom", "htap", "country"] = "default"
    map_zoom: str = "World"
    add_model_maps: bool = False
    modelorder_from_config: bool = True
    obsorder_from_config: bool = True
    var_order_menu: tuple[str, ...] = ()
    obs_order_menu: tuple[str, ...] = ()
    stats_order_menu: tuple[str, ...] = ()
    model_order_menu: tuple[str, ...] = ()
    hide_charts: tuple[str, ...] = ()
    hide_pages: tuple[str, ...] = ()
    ts_annotations: dict[str, str] = Field(default_factory=dict)
    pages: tuple[str, ...] = ("maps", "evaluation", "intercomp", "overall", "infos")



from pathlib import Path
import string

import numpy as np
import pytest

from pyaerocom import ungriddeddata
from pyaerocom.exceptions import DataCoverageError, VariableDefinitionError
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.ungriddeddata_structured import UngriddedDataStructured
from tests.fixtures.stations import FAKE_STATION_DATA


@pytest.fixture(scope="module")
def ungridded_empty():
    return UngriddedDataStructured()


@pytest.fixture(scope="function")
def aeronetsunv3lev2_subset_uds(aeronetsunv3lev2_subset):
    uds = UngriddedDataStructured()
    # convert ungriddeddata to ungriddeddata_structured by merging
    uds.merge(aeronetsunv3lev2_subset, new_obj=False)
    assert not uds.is_empty
    return uds


def test_ungridded_new():
    ud = UngriddedDataStructured(num_points=1000)
    assert ud._dra._capacity == 1000
    assert len(ud._dra._array["meta_id"]) == 1000
    assert len(ud._dra.data["meta_id"]) == 0


def test_coordinate_access():
    d = UngriddedDataStructured()

    stat_names = list(string.ascii_lowercase)
    lons = np.arange(len(stat_names))
    lats = np.arange(len(stat_names)) - 90
    alts = np.arange(len(stat_names)) * 13

    for i, n in enumerate(stat_names):
        d.metadata[i] = dict(
            data_id="testcase",
            station_name=n,
            latitude=lats[i],
            longitude=lons[i],
            altitude=alts[i],
        )

    assert d.station_name == stat_names
    assert all(d.latitude == lats)
    assert all(d.longitude == lons)
    assert all(d.altitude == alts)

    with pytest.raises(DataCoverageError):
        d.to_station_data("a")

    c = d.station_coordinates
    assert c["station_name"] == stat_names
    assert all(c["latitude"] == lats)
    assert all(c["longitude"] == lons)
    assert all(c["altitude"] == alts)


# sites in aeronet data

ALL_SITES = [
    "AAOT",
    "ARIAKE_TOWER",
    "Agoufou",
    "Alta_Floresta",
    "American_Samoa",
    "Amsterdam_Island",
    "Anmyon",
    "Avignon",
    "Azores",
    "BORDEAUX",
    "Barbados",
    "Blyth_NOAH",
    "La_Paz",
    "Mauna_Loa",
    "Tahiti",
    "Taihu",
    "Taipei_CWB",
    "Tamanrasset_INM",
    "The_Hague",
    "Thessaloniki",
    "Thornton_C-power",
    "Trelew",
]


@pytest.mark.parametrize(
    "args,sitenames",
    [
        ({"station_name": ["Tr*", "Mauna*"]}, ["Trelew", "Mauna_Loa"]),
        (
            {"station_name": ["Tr*", "Mauna*"], "negate": "station_name"},
            [x for x in ALL_SITES if x not in ["Trelew", "Mauna_Loa"]],
        ),
        (
            {"altitude": [0, 1000], "negate": "altitude"},
            ["La_Paz", "Mauna_Loa", "Tamanrasset_INM"],
        ),
        ({"station_name": "Tr*"}, ["Trelew"]),
        (
            {"station_name": "Tr*", "negate": "station_name"},
            [x for x in ALL_SITES if not x == "Trelew"],
        ),
    ],
)
def test_filter_by_meta(aeronetsunv3lev2_subset_uds, args, sitenames):
    data = aeronetsunv3lev2_subset_uds
    assert isinstance(data, UngriddedDataStructured)
    assert data.get_data_revision("AeronetSunV3L2Subset.daily") == "n/d"
    subset = data.filter_by_meta(**args)
    assert subset.get_data_revision("AeronetSunV3L2Subset.daily") == "n/d"
    sites = [x["station_name"] for x in subset.metadata.values()]
    stats = sorted(list(dict.fromkeys(sites)))
    assert sorted(sitenames) == stats


def test_ebas_revision(data_scat_jungfraujoch: UngriddedDataContainer):
    assert isinstance(data_scat_jungfraujoch, UngriddedDataStructured)
    assert data_scat_jungfraujoch.get_data_revision("EBASSubset") == "20220101"


def test_cache_reload(data_scat_jungfraujoch: UngriddedDataContainer, tmp_path: Path):
    path = tmp_path / "ungridded_scat_jungfraujoch.pkl"
    file = data_scat_jungfraujoch.save_as(file_name=path.name, save_dir=path.parent)
    assert Path(file) == path
    assert path.exists()
    data = UngriddedDataContainer.from_cache(data_dir=path.parent, file_name=path.name)
    assert data.shape == data_scat_jungfraujoch.shape


def test_check_unit(data_scat_jungfraujoch):
    data_scat_jungfraujoch.check_unit("sc550aer", unit="1/Mm")
    from pyaerocom.exceptions import MetaDataError

    with pytest.raises(MetaDataError):
        data_scat_jungfraujoch.check_unit("sc550aer", unit="m-1")


def test_from_single_station_data():
    stat = FAKE_STATION_DATA["station_data1"]
    d = ungriddeddata.UngriddedData.from_station_data(stat)
    data0 = stat.ec550aer
    data1 = d.all_datapoints_var("ec550aer")
    assert data0 == pytest.approx(data1, abs=1e-20)


def test_has_flag_data(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    assert isinstance(aeronetsunv3lev2_subset_uds.has_flag_data, np.bool_ | bool)


def test_is_filtered(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    assert isinstance(aeronetsunv3lev2_subset_uds.is_filtered, np.bool_ | bool)


def test_available_meta_keys(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    assert isinstance(aeronetsunv3lev2_subset_uds.available_meta_keys, list)
    assert all(isinstance(key, str) for key in aeronetsunv3lev2_subset_uds.available_meta_keys)


def test_nonunique_station_names(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    assert isinstance(aeronetsunv3lev2_subset_uds.nonunique_station_names, list)


def test_set_flags_nan_error(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    data = aeronetsunv3lev2_subset_uds.copy()
    with pytest.raises(AttributeError):
        data = data.data.set_flags_nan(inplace=True)


def test_remove_outliers(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    data = aeronetsunv3lev2_subset_uds.copy()
    assert not data.filter_hist
    new = data.remove_outliers(var_name="od550aer", low=0, high=0)
    assert new.filter_hist


def test_extract_var(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    data = aeronetsunv3lev2_subset_uds.copy()
    od = data.extract_var("od550aer")
    assert not data.is_filtered
    assert od.is_filtered
    assert od.shape[0] < data.shape[0]


def test_extract_var_error(aeronetsunv3lev2_subset_uds: UngriddedDataStructured):
    data = aeronetsunv3lev2_subset_uds.copy()
    with pytest.raises(VariableDefinitionError):
        data.extract_var("nope")



"""
Small helper utility functions for pyaerocom
"""

import abc
import logging
import os
from collections.abc import MutableMapping
from pathlib import Path
from typing import TypeVar

import numpy as np
from typing_extensions import TypedDict

from pyaerocom._warnings import ignore_warnings

logger = logging.getLogger(__name__)


def invalid_input_err_str(argname, argval, argopts):
    """Just a small helper to format an input error string for functions

    Parameters
    ----------
    argname : str
        name of input argument
    argval
        (invalid) value of input argument
    argopts
        possible input args for arg

    Returns
    -------
    str
        formatted string that can be parsed to an Exception
    """

    return f"Invalid input for {argname} ({argval}), choose from {argopts}"


def check_dir_access(path):
    """Uses multiprocessing approach to check if location can be accessed

    Parameters
    ----------
    loc : str
        path that is supposed to be checked

    Returns
    -------
    bool
        True, if location is accessible, else False
    """
    if not isinstance(path, str):
        return False

    return os.access(path, os.R_OK)


def check_write_access(path):
    """Check if input location provides write access

    Parameters
    ----------
    path : str
        directory to be tested

    """
    if not isinstance(path, str):
        # not a path
        return False

    return os.access(path, os.W_OK)


def _class_name(obj):
    """Returns class name of an object"""
    return type(obj).__name__


# TODO: Check to see if instances of these classes can instead use pydantic
class Validator(abc.ABC):
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, obj, objtype=None):
        try:
            return obj.__dict__[self._name]
        except (AttributeError, KeyError):
            raise AttributeError("value not set...")

    def __set__(self, obj, val):
        val = self.validate(val)
        obj.__dict__[self._name] = val

    @abc.abstractmethod
    def validate(self, val):
        pass


class TypeValidator(Validator):
    def __init__(self, type):
        self._type = type

    def validate(self, val):
        if not isinstance(val, self._type):
            raise ValueError(f"need instance of {self._type}")
        return val


class StrType(Validator):
    def validate(self, val):
        if not isinstance(val, str):
            raise ValueError(f"need str, got {val}")
        return val


class Loc(abc.ABC):
    """Abstract descriptor representing a path location

    Descriptor: TODO
    See here: https://docs.python.org/3/howto/descriptor.html#complete-practical-example

    Note
    ----
    - Child classes need to implement :func:`create`
    - value is allowed to be `None` in which case no checks are performed
    """

    def __init__(
        self,
        default=None,
        assert_exists=False,
        auto_create=False,
        tooltip=None,
    ):
        self.assert_exists = assert_exists
        self.auto_create = auto_create
        self.tooltip = "" if tooltip is None else tooltip
        self.__set__(self, default)

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, objtype=None):
        try:
            val = obj.__dict__[self.name]
        except (KeyError, AttributeError):
            val = self.default
        return val

    def __set__(self, obj, value):
        value = self.validate(value)
        try:
            obj.__dict__[self.name] = value
        except AttributeError:
            self.default = value

    def validate(self, value):
        if value is None:
            return value
        elif isinstance(value, Path):
            value = str(value)
        if not isinstance(value, str):
            raise ValueError(value)
        if self.assert_exists and not os.path.exists(value):
            if self.auto_create:
                self.create(value)
            else:
                raise FileNotFoundError(value)
        return value

    @abc.abstractmethod
    def create(self, value):
        pass


class DirLoc(Loc):
    def create(self, value):
        os.makedirs(value, exist_ok=True)
        logger.info(f"created directory {value}")


class AsciiFileLoc(Loc):
    def create(self, value):
        logger.info(f"create ascii file {value}")
        open(value, "w").close()


class BrowseDict(MutableMapping):
    """Dictionary-like object with getattr and setattr options

    Extended dictionary that supports dynamic value generation (i.e. if an
    assigned value is callable, it will be executed on demand).
    """

    ADD_GLOB = []
    FORBIDDEN_KEYS = []
    #: Keys to be ignored when converting to json
    IGNORE_JSON = []
    MAXLEN_KEYS = 1e2
    SETTER_CONVERT = {}

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    @property
    def _class_name(self):
        return _class_name(self)

    def keys(self):
        return list(self.__dict__) + self.ADD_GLOB

    def _get_glob_vals(self):
        return [getattr(self, x) for x in self.ADD_GLOB]

    def values(self):
        return [getattr(self, x) for x in self.keys()]

    def items(self):
        for key in self.keys():
            yield key, getattr(self, key)

    def __setitem__(self, key, val) -> None:
        key, val, ok = self._setitem_checker(key, val)
        if not ok:
            return
        if bool(self.SETTER_CONVERT):
            for fromtp, totp in self.SETTER_CONVERT.items():
                if isinstance(val, fromtp):
                    if fromtp is dict:
                        val = totp(**val)
                    else:
                        val = totp(val)

        if isinstance(key, str):
            if len(key) > self.MAXLEN_KEYS:
                raise KeyError(f"key {key} exceeds max length of {self.MAXLEN_KEYS}")
            if key in self.FORBIDDEN_KEYS:
                raise KeyError(f"invalid key {key}")
        setattr(self, key, val)

    def _setitem_checker(self, key, val):
        return key, val, True

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except TypeError:
            # if key is not str
            return self.__dict__[key]
        except AttributeError as e:
            raise KeyError(e)

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        """echoes class, id, & reproducible representation in the REPL"""
        _repr = repr(self.__dict__)
        return f"{_class_name(self)}: {_repr}"

    def to_dict(self):
        out = {}
        for key, val in self.items():
            out[key] = val
        return out

    def json_repr(self) -> dict:
        """
        Convert object to serializable json dict

        Returns
        -------
        dict
            content of class

        """
        output = {}
        for key, val in self.items():
            if key in self.IGNORE_JSON:
                continue
            if hasattr(val, "json_repr"):
                val = val.json_repr()
            output[key] = val
        return output

    def import_from(self, other) -> None:
        """
        Import key value pairs from other object

        Other than :func:`update` this method will silently ignore input
        keys that are not contained in this object.

        Parameters
        ----------
        other : dict or BrowseDict
            other dict-like object containing content to be updated.

        Raises
        ------
        ValueError
            If input is invalid type.

        Returns
        -------
        None

        """
        if not isinstance(other, dict | BrowseDict):
            raise ValueError("need dict-like object")
        for key, val in other.items():
            if key in self:
                self[key] = val
            elif key in self.FORBIDDEN_KEYS:
                raise KeyError(f"invalid key {key}")

    def pretty_str(self):
        return dict_to_str(self.to_dict())

    def __str__(self):
        return str(self.to_dict())


class ConstrainedContainer(BrowseDict):
    """Restrictive dict-like class with fixed keys

    This class enables to create dict-like objects that have a fixed set of
    keys and value types (once assigned). Optional values may be instantiated
    as None, in which case the first time instantiation defines its type.

    Note
    ----
    The limitations for assignments are only restricted to setitem operations
    and attr assignment via "." works like in every other class.

    Example
    -------
    >>> class MyContainer(ConstrainedContainer):
    ...    def __init__(self):
    ...        self.val1 = 1
    ...        self.val2 = 2
    ...        self.option = None
    >>> mc = MyContainer()
    >>> mc['option'] = 42
    """

    CRASH_ON_INVALID = True

    def __setitem__(self, key, val):
        super().__setitem__(key, val)

    def _invoke_dtype(self, current_tp, val):
        return current_tp(**val)

    def _check_valtype(self, key, val):
        current_tp = type(self[key])
        if type(val) is not current_tp and isinstance(self[key], BrowseDict):
            val = current_tp(**val)
        return val

    def _setitem_checker(self, key, val):
        """make sure no new attr is added

        Note
        ----
        Only used in __setitem__ not in __setattr__.
        """
        if key not in dir(self):
            if self.CRASH_ON_INVALID:
                raise ValueError(f"Invalid key {key}")
            logger.warning(f"Invalid key {key} in {self._class_name}. Will be ignored.")
            return key, val, False

        current = getattr(self, key)
        val = self._check_valtype(key, val)
        current_tp = type(current)

        if current is not None and not isinstance(val, current_tp):
            raise ValueError(
                f"Invalid type {type(val)} for key: {key}. Need {current_tp} "
                f"(Current value: {current})"
            )
        return key, val, True


class NestedContainer(BrowseDict):
    def _occurs_in(self, key) -> list:
        objs = []
        if key in self:
            objs.append(self)
        for k, v in self.items():
            if isinstance(v, dict | BrowseDict) and key in v:
                objs.append(v)
            if len(objs) > 1:
                print(key, "is contained in multiple containers ", objs)
        return objs

    def keys_unnested(self) -> list:
        keys = []
        for key, val in self.items():
            keys.append(key)
            if isinstance(val, NestedContainer):
                keys.extend(val.keys_unnested())
            elif isinstance(val, ConstrainedContainer | dict):
                for subkey, subval in val.items():
                    keys.append(subkey)
        return keys

    def update(self, **settings):
        for key, val in settings.items():
            to_update = self._occurs_in(key)
            if len(to_update) == 0:
                raise AttributeError(f"invalid key {key}")
            for obj in to_update:
                obj[key] = val

    def __str__(self):
        return dict_to_str(self)


def merge_dicts(dict1, dict2, discard_failing=True):
    """Merge two dictionaries

    Parameters
    ----------
    dict1 : dict
        first dictionary
    dict2 : dict
        second dictionary
    discard_failing : bool
        if True, any key, value pair that cannot be merged from the 2nd into
        the first will be skipped, which means, the value of the output dict
        for that key will be the one of the first input dict. All keys that
        could not be merged can be accessed via key 'merge_failed' in output
        dict. If False, any Exceptions that may occur will be raised.

    Returns
    -------
    dict
        merged dictionary
    """
    # make a copy of the first dictionary
    new = dict(**dict1)
    merge_failed = []
    # loop over all entries of second one
    for key, val in dict2.items():
        try:
            # entry does not exist in first dict or is None
            if key not in new or new[key] is None:
                new[key] = val
                continue
            # get value of first input dict
            this = new[key]

            # check if values are the same and skip (try/except is because for
            # some data types equality tests may return iterable (e.g. compare
            # 2 numpy arrays))
            try:
                if this == val:
                    continue
            except Exception:
                try:
                    if (this == val).all():
                        continue
                except Exception:
                    pass

            # both values are strings, merge with ';' delim
            if isinstance(this, str) and isinstance(val, str):
                new[key] = f"{this};{val}"

            elif isinstance(this, list) and isinstance(val, list):
                for item in val:
                    if item not in this:
                        this.append(item)
                new[key] = this

            elif all(isinstance(x, dict) for x in (this, val)):
                new[key] = merge_dicts(this, val)

            elif any(isinstance(x, list) for x in (this, val)):
                if isinstance(this, list):
                    lst = this
                    check = val  # this is not list
                else:
                    lst = val
                    check = this  # this is not list
                for item in lst:
                    if type(item) is not type(check):
                        raise ValueError(
                            f"Cannot merge key {key} since items in {lst} "
                            f"are of different type, that does not match {check}"
                        )
                lst.append(check)
                new[key] = lst

            else:
                new[key] = [this, val]
        except Exception:
            if discard_failing:
                merge_failed.append(key)
            else:
                raise
    new["merge_failed"] = merge_failed

    return new


def chk_make_subdir(base, name):
    """Check if sub-directory exists in parent directory"""
    d = os.path.join(base, name)
    os.makedirs(d, exist_ok=True)
    return d


def list_to_shortstr(lst, indent=0):
    """Custom function to convert a list into a short string representation"""

    def _short_lst_fmt(lin):
        lout = []
        for val in lin:
            try:
                with ignore_warnings(
                    RuntimeWarning,
                    "divide by zero encountered in log10",
                    "overflow encountered in scalar multiply",
                ):
                    ndigits = -1 * np.floor(np.log10(abs(np.asarray(val)))).astype(int) + 2
                lout.append(f"{val:.{ndigits}f}")
            except Exception:
                lout.append(val)
        return lout

    name_str = f"{type(lst).__name__} ({len(lst)} items): "
    indentstr = indent * " "
    if len(lst) == 0:
        return f"{indentstr}{name_str}[]"
    elif len(lst) < 6:
        lfmt = _short_lst_fmt(lst)
        return f"{indentstr}{name_str}{lfmt}"
    else:  # first 2 and last 2 items
        lfmt = _short_lst_fmt([lst[0], lst[1], lst[-2], lst[-1]])
        s = f"{indentstr}{name_str}[{lfmt[0]}, {lfmt[1]}, ..., {lfmt[2]}, {lfmt[3]}]"

    return s


T = TypeVar("T")


def sort_dict_by_name(d: dict[str, T], pref_list: list[str] | None = None) -> dict[str, T]:
    """Sort entries of input dictionary by their names and return ordered

    Parameters
    ----------
    d : dict
        input dictionary
    pref_list : list, optional
        preferred order of items (may be subset of keys in input dict)

    Returns
    -------
    dict
        sorted and ordered dictionary
    """
    if pref_list is None:
        pref_list = []
    s = {}
    sorted_keys = sorted(d)
    for k in pref_list:
        if k in d:
            s[k] = d[k]
    for k in sorted_keys:
        if k not in pref_list:
            s[k] = d[k]
    return s


def dict_to_str(dictionary, indent=0, ignore_null=False):
    """Custom function to convert dictionary into string (e.g. for print)

    Parameters
    ----------
    dictionary : dict
        the dictionary
    indent : int
        indent of dictionary content
    ignore_null : bool
        if True, None entries in dictionary are ignored

    Returns
    -------
    str
        the modified input string

    """
    if len(dictionary) == 0:
        return "{}"
    elif len(dictionary) == 1:
        pre = ind = offs = ""
    else:
        pre = "\n"
        ind = indent * " "
        offs = " "
    s = "{"

    for key, val in dictionary.items():
        if ignore_null and val is None:
            continue
        elif isinstance(val, dict | BrowseDict):
            val = dict_to_str(val, indent + 2)
        elif isinstance(val, list):
            val = list_to_shortstr(val, indent=indent)
        elif isinstance(val, np.ndarray) and val.ndim == 1:
            val = list_to_shortstr(val, indent=indent)
        s += f"{pre}{ind}{offs}{key}: {val}"
    s += pre + ind + "}"
    return s


def str_underline(title: str, indent: int = 0):
    """Create underlined string"""
    length = indent + len(title)
    underline = "-" * len(title)
    return f"{title:>{length}}\n{underline:>{length}}"


class RegridResDeg(TypedDict):
    """Typed dict for regridding resolution degrees"""

    lat_res_deg: float
    lon_res_deg: float


class LayerLimits(TypedDict):
    """Typed dict of 3D colocation layer limits"""

    start: float
    end: float



"""
Caching class for reading and writing of ungridded data Cache objects
"""

import glob
import logging
import os
import pickle
from collections.abc import Iterator
from pathlib import Path

from pyaerocom import const
from pyaerocom.exceptions import CacheReadError, CacheWriteError
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.ungriddeddata_structured import UngriddedDataStructured
from pyaerocom.io.pyaro.read_pyaro import ReadPyaro

logger = logging.getLogger(__name__)


# TODO: Write data attribute list contains_vars in header of pickled file and
# check if variables match the request
class CacheHandlerUngridded:
    """Interface for reading and writing of cache files

    Cache filename mask is

    <data_id>_<var>.pkl

    e.g. EBASMC_scatc550aer.pkl

    Attributes
    ----------
    reader : ReadUngriddedBase
        reading class for dataset
    loaded_data : dict
        dictionary containing successfully loaded instances of single variable
        :class:`UngriddedData` objects (keys are variable names)
    """

    __version__ = "1.13"
    #: Cache file header keys that are checked (and required unchanged) when
    #: reading a cache file
    CACHE_HEAD_KEYS = [
        "pyaerocom_version",
        "newest_file_in_read_dir",
        "newest_file_date_in_read_dir",
        "data_revision",
        "reader_version",
        "ungridded_data_class",
        "ungridded_data_version",
        "cacher_version",
    ]

    _cachable_classes = {x.__name__: x for x in (UngriddedDataStructured, UngriddedData)}

    def __init__(self, reader=None, cache_dir=None, **kwargs):
        self._reader = None
        if reader is not None:
            self._reader = reader

        self.loaded_data = {}
        self._cache_dir = cache_dir

    @property
    def reader(self):
        """Instance of reader class"""
        if self._reader is None:
            raise AttributeError("No reader class assigned to cache object")
        return self._reader

    @reader.setter
    def reader(self, val):
        from pyaerocom.io import ReadUngriddedBase

        if not isinstance(val, ReadUngriddedBase):
            raise TypeError("Invalid input for reader")
        self._reader = val
        self.loaded_data = {}

    @property
    def cache_dir(self):
        """Directory where cache data objects are stored"""
        if self._cache_dir is not None:
            return self._cache_dir
        return const.CACHEDIR

    @cache_dir.setter
    def cache_dir(self, val):
        if not isinstance(val, str) or not os.path.exists(val):
            raise FileNotFoundError(f"Input directory does not exist: {val}")
        self._cache_dir = val

    @property
    def data_id(self):
        """Data ID of the associated dataset"""
        return self.reader.data_id

    @property
    def src_data_dir(self):
        """Data source directory of the associated dataset

        Needed to check whether an existing cache file is outdated
        """
        return self.reader.data_dir

    def default_file_name(self, var_name):
        """File name of cache file


        Parameters
        ----------
        var_name : str
            name of variable to be cached.

        Returns
        -------
        str
            file name of pickle file
        """
        name = "_".join([self.data_id, var_name])
        return name + ".pkl"

    def file_path(self, var_or_file_name, cache_dir=None):
        """File path of cache file

        Parameters
        ----------
        var_or_file_name : str
            name of output filename or variable that is supposed to be stored.
            Default usage is to provide variable and then
            :func:`default_file_name` is used. Can be None if
            input `data` contains only a single variable.
        cache_dir : str, optional
            output directory (default is pyaerocom cache dir accessed via
            :func:`cache_dir`).

        Returns
        -------
        str
            output file path
        """
        if not var_or_file_name.endswith(".pkl"):
            var_or_file_name = self.default_file_name(var_or_file_name)
        if cache_dir is None:
            cache_dir = self.cache_dir
        if cache_dir is None:
            raise FileNotFoundError("Specified output directory is None")
        if not os.path.exists(cache_dir):
            raise FileNotFoundError(f"Specified output directory does not exist:{cache_dir}")
        return os.path.join(cache_dir, var_or_file_name)

    def _check_pkl_head_vs_database(self, head):
        current = self._cache_meta_info(head["ungridded_data_class"])

        if not isinstance(head, dict):
            raise CacheReadError("Invalid cache file")
        pya_version = "pyaerocom_version"
        for k, v in head.items():
            if k not in current:
                raise CacheReadError(f"Invalid cache header key: {k}")
            else:
                if k == pya_version:
                    # ignore pya_version
                    continue
                elif v != current[k]:
                    logger.info(f"{k} is outdated (value: {v}). Current value: {current[k]}")
                    return False
        if current[pya_version] != head[pya_version]:
            logger.info(
                f"cache generated with pyaerocom {head[pya_version]}, current is {current[pya_version]}"
            )
        return True

    def _cache_meta_info(self, data_classname):
        """Dictionary containing relevant caching meta-info

        Parameters
        ----------
        data_class : implementation class of UngriddedDataContainer

        """
        dataclass = self._cachable_classes[data_classname]
        try:
            newestp = max(glob.iglob(os.path.join(self.src_data_dir, "*")), key=os.path.getmtime)
            newest_date = os.path.getmtime(newestp)
            newest = os.path.basename(newestp)

        except Exception:
            newest = None
            newest_date = None

        try:
            rev = self.reader.data_revision
            reader_ver = self.reader.__version__
        except AttributeError:
            rev = None
            reader_ver = None
        current = dict.fromkeys(self.CACHE_HEAD_KEYS)
        from pyaerocom import __version__

        current["pyaerocom_version"] = __version__
        current["newest_file_in_read_dir"] = newest
        current["newest_file_date_in_read_dir"] = newest_date
        current["data_revision"] = rev
        current["reader_version"] = reader_ver
        current["ungridded_data_class"] = data_classname
        current["ungridded_data_version"] = dataclass.__version__
        current["cacher_version"] = self.__version__
        if isinstance(self._reader, ReadPyaro):
            current["pyaro_config"] = self.reader.config
        return current

    def check_and_load(self, var_or_file_name, force_use_outdated=False, cache_dir=None):
        """Check if cache file exists and load

        Note
        ----
        If a cache file exists for this database, but cannot be loaded or is
        outdated against pyaerocom updates, then it will be removed (the latter
        only if :attr:`pyaerocom.const.RM_CACHE_OUTDATED` is True).

        Parameters
        ----------
        var_or_file_name : str
            name of output filename or variable that is supposed to be stored.
            Default usage is to provide variable and then
            :func:`default_file_name` is used. Can be None if
            input `data` contains only a single variable.ead
        force_use_outdated : bool
            if True, read existing cache file even if it is not up to date or
            pyaerocom version changed (not recommended to use)
        cache_dir : str, optional
            output directory (default is pyaerocom cache dir accessed via
            :func:`cache_dir`).

        Returns
        -------
        bool
            True, if cache file exists and could be successfully loaded, else
            False. Note: if import is successful, the corresponding data object
            (instance of :class:`pyaerocom.UngriddedData` can be accessed via
            :attr:`loaded_data'

        Raises
        ------
        TypeError
            if cached file is not an instance of :class:`pyaerocom.UngriddedData`
            class (which should not happen)
        """
        try:
            fp = self.file_path(var_or_file_name, cache_dir=cache_dir)
        except FileNotFoundError as e:
            logger.warning(repr(e))
            return False

        if not os.path.isfile(fp):
            logger.info(f"Cache file does not exist: {fp}")
            return False

        delete_existing = const.RM_CACHE_OUTDATED if not force_use_outdated else False

        with open(fp, "rb") as in_handle:
            last_meta = pickle.load(in_handle)
            if "ungridded_data_class" not in last_meta:
                # backward compatibility < 0.29
                last_meta["ungridded_data_class"] = UngriddedData.__name__

            if force_use_outdated:
                ok = True
            else:
                try:
                    ok = self._check_pkl_head_vs_database(last_meta)
                except Exception as e:
                    ok = False
                    delete_existing = True
                    logger.exception(
                        f"File error in cached data file {fp}. "
                        f"File will be removed and data reloaded. Error: {repr(e)}"
                    )
            if ok:
                # everything is okay, or forced
                data = pickle.load(in_handle)

        if not ok:
            # Delete the cache file if it is outdated, after handle is closed
            logger.info(
                f"Aborting reading cache file {fp}. Aerocom database "
                f"or pyaerocom version has changed compared to cached version"
            )
            if delete_existing:  # something was wrong
                logger.info(f"Deleting outdated cache file: {fp}")
                os.remove(fp)
            return False

        if not isinstance(data, UngriddedDataContainer):
            raise TypeError(
                f"Unexpected data type stored in cache file, need instance of UngriddedData, "
                f"got {type(data)}"
            )

        self.loaded_data[var_or_file_name] = data
        logger.info(f"Successfully loaded cache file {fp} as {type(data).__name__}")
        return True

    def write(self, data: UngriddedDataContainer, var_or_file_name=None, cache_dir=None):
        """Write single-variable instance of UngriddedData to cache

        Parameters
        ----------
        data : UngriddedDataContainer
            object containing the data (possibly containing multiple variables)
        var_or_file_name : str, optional
            name of output filename or variable that is supposed to be stored.
            Default usage is to provide variable and then
            :func:`default_file_name` is used. Can be None if
            input `data` contains only a single variable.
        cache_dir : str, optional
            output directory (default is pyaerocom cache dir accessed via
            :func:`cache_dir`).

        Returns
        -------
        str
            output file path
        """
        if not isinstance(data, UngriddedDataContainer):
            raise TypeError(
                f"Invalid input, need instance of UngriddedDataContainer, got {type(data)}"
            )

        meta = self._cache_meta_info(type(data).__name__)

        if not var_or_file_name.endswith(".pkl"):
            var_name = var_or_file_name
            if len(data.contains_datasets) > 1:
                raise CacheWriteError(
                    f"Input UngriddedDataContainer object contains datasets: {data.contains_datasets}. "
                    f"Can only write single dataset objects"
                )
            if var_name is None:
                if len(data.contains_vars) > 1:
                    raise CacheWriteError(
                        f"Input UngriddedDataContainer object for {self.reader.data_id} "
                        f"contains more than one variable: {data.contains_vars}. "
                        f"Please specify which variable should be cached"
                    )
                var_name = data.contains_vars[0]

            elif var_name not in data.contains_vars:
                raise CacheWriteError(
                    f"Cannot write cache file: variable {var_name} "
                    f"does not exist in input UngriddedDataContainer object"
                )

            if len(data.contains_vars) > 1:
                data = data.extract_var(var_name)

        fp = self.file_path(var_or_file_name, cache_dir=cache_dir)
        logger.info(f"Writing cache file: {fp}")
        success = True
        # OutHandle = gzip.open(c__cache_file, 'wb') # takes too much time
        out_handle = open(fp, "wb")

        try:
            # write cache header
            pickle.dump(meta, out_handle, pickle.HIGHEST_PROTOCOL)
            # write data
            pickle.dump(data, out_handle, pickle.HIGHEST_PROTOCOL)

        except Exception as e:
            logger.exception(f"Failed to write cache: {repr(e)}")
            success = False
        finally:
            out_handle.close()
            if not success:
                os.remove(fp)
        logger.info(f"Wrote: {fp}")
        return fp

    def __str__(self):
        return f"pyaerocom.CacheHandlerUngridded\nDefault cache dir: {self.cache_dir}"


def list_cache_files() -> Iterator[Path]:
    """
    List all pickled data objects in cache directory

    If not set differently, the cache directory is the pyaerocom default,
    accessible via :attr:`pyaerocom.const.CACHEDIR`.

    """
    ch = CacheHandlerUngridded()
    return Path(ch.cache_dir).glob("*.pkl")



import fnmatch
import logging
import os
import re

from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.exceptions import DataSearchError

logger = logging.getLogger(__name__)


class AerocomBrowser(BrowseDict):
    """Interface for browsing all Aerocom data directories

    Note
    ----
    Use :func:`browse` to find directories matching a
    certain search pattern.
    The class methods :func:`find_matches` and :func:`find_dir` both use
    :func:`browse`, the only difference is, that the :func:`find_matches` adds
    the search result (a list with strings) to

    """

    def _browse(self, name_or_pattern, ignorecase=True, return_if_match=True):
        """Search all Aerocom data directories that match input name or pattern

        Note
        ----
        Please do not use this function but either
        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored
        return_if_match : bool
            if True, then the data directory is returned as string, if it can
            be found, else, only a list is returned that contains all
            matches. The latter takes longer since the whole database is
            searched.

        Returns
        -------
        :obj:`str` or :obj:`list`
            Data directory (str, if ``return_if_match`` is True) or list
            containing valid Aerocom names (which can then be used to
            retrieve the paths)

        Raises
        ------
        DataSearchError
            if no match or no unique match can be found
        """
        pattern = fnmatch.translate(name_or_pattern)
        _candidates = []
        _msgs = []
        _warnings = []

        for obs_id, obs_path in const.OBSLOCS_UNGRIDDED.items():
            if ignorecase:
                match = name_or_pattern.lower() == obs_id.lower()
            else:
                match = name_or_pattern == obs_id
            if match:
                logger.info(f"Found match for search pattern in obs network directories {obs_id}")
                path = os.path.normpath(obs_path)
                if os.path.exists(path):
                    self[obs_id] = path
                    _candidates.append(obs_id)
                    if return_if_match:
                        return path
            else:
                if ignorecase:
                    match = bool(re.search(pattern, obs_id, re.IGNORECASE))
                else:
                    match = bool(re.search(pattern, obs_id))
                if match:
                    path = os.path.normpath(obs_path)
                    if os.path.exists(path):
                        self[obs_id] = path
                        _candidates.append(obs_id)
                        if return_if_match:
                            return path

        for search_dir in const.DATA_SEARCH_DIRS:
            # get the directories
            if os.path.isdir(search_dir):
                subdirs = [
                    x for x in os.listdir(search_dir) if os.path.isdir(os.path.join(search_dir, x))
                ]
                for subdir in subdirs:
                    if ignorecase:
                        match = bool(re.search(pattern, subdir, re.IGNORECASE))
                    else:
                        match = bool(re.search(pattern, subdir))
                    if match:
                        _dir = os.path.normpath(os.path.join(search_dir, subdir))
                        _rnsubdir = os.path.join(_dir, "renamed")
                        if os.path.isdir(_rnsubdir):
                            logger.info(f"{_dir} has subdir renamed. Using that one")
                            _dir = _rnsubdir
                        if any([_dir in x for x in self.values()]):
                            # directory was already found before
                            continue
                        # append name of candidate ...
                        _candidates.append(subdir)
                        # ... and the corresponding data directory
                        self[subdir] = _dir

                        # now check if it is actually an exact match, if
                        # applicable
                        if return_if_match:
                            if ignorecase:
                                match = name_or_pattern.lower() == subdir.lower()
                            else:
                                match = name_or_pattern == subdir
                            if match:
                                logger.info(f"Found match for ID {name_or_pattern}")
                                if return_if_match:
                                    return _dir

            else:
                _msgs.append(f"directory {search_dir} does not exist\n")
        for msg in _msgs:
            logger.info(msg)

        for warning in _warnings:
            logger.warning(warning)

        if len(_candidates) == 0:
            raise DataSearchError(
                f"No matches could be found for search pattern {name_or_pattern}"
            )
        if return_if_match:
            if len(_candidates) == 1:
                logger.info(
                    f"Found exactly one match for search pattern "
                    f"{name_or_pattern}: {_candidates[0]}"
                )
                return self[_candidates[0]]
            raise DataSearchError(
                f"Found multiple matches for search pattern {name_or_pattern}. "
                f"Please choose from {_candidates}"
            )
        return _candidates

    @property
    def dirs_found(self):
        """All directories that were found"""
        return list(self.values())

    @property
    def ids_found(self):
        """All data IDs that were found"""
        return list(self)

    def find_data_dir(self, name_or_pattern, ignorecase=True):
        """Find match of input name or pattern in Aerocom database

        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored

        Returns
        -------
        str
            data directory of match

        Raises
        ------
        DataSearchError
            if no matches or no unique match can be found
        """
        if name_or_pattern in self:
            logger.info(f"{name_or_pattern} found in instance of AerocomBrowser")
            return self[name_or_pattern]
        logger.info(f"Searching database for {name_or_pattern}")
        return self._browse(
            name_or_pattern, ignorecase=ignorecase, return_if_match=True
        )  # returns list

    def find_matches(self, name_or_pattern, ignorecase=True):
        """Search all Aerocom data directories that match input name or pattern

        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored

        Returns
        -------
        list
            list of names that match the pattern (corresponding paths can be
            accessed from this class instance)

        Raises
        ------
        DataSearchError
            if no matches can be found
        """
        return self._browse(
            name_or_pattern, ignorecase=ignorecase, return_if_match=False
        )  # returns list



"""
Classes and methods to perform high-level colocation.
"""

import glob
import logging
import os
import traceback
import warnings
from collections import defaultdict
from typing import Any

import pandas as pd

from pyaerocom import const
from pyaerocom.colocation.colocation_utils import (
    colocate_gridded_gridded,
    colocate_gridded_ungridded,
    correct_model_stp_coldata,
)
from pyaerocom.exceptions import (
    ColocationError,
    ColocationSetupError,
    DataCoverageError,
)
from pyaerocom.griddeddata import GriddedData
from pyaerocom.helpers import start_stop, to_datestring_YYYYMMDD
from pyaerocom.io import ReadCAMS2_83, ReadGridded, ReadUngridded
from pyaerocom.io.helpers import get_all_supported_ids_ungridded
from pyaerocom.io.mscw_ctm.reader import ReadMscwCtm
from pyaerocom.stats.mda8.const import MDA8_INPUT_VARS
from pyaerocom.stats.mda8.mda8 import mda8_colocated_data
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.units import Unit
from pyaerocom.units.datetime import get_lowest_resolution, to_pandas_timestamp
from pyaerocom.units.harmonise import harmonise_units

from .colocated_data import ColocatedData
from .colocation_3d import ColocatedDataLists, colocate_vertical_profile_gridded
from .colocation_setup import ColocationSetup

logger = logging.getLogger(__name__)


class Colocator:
    """High level class for running co-location

    Note
    ----
    This object requires and instance from :class:`ColocationSetup`.
    """

    SUPPORTED_GRIDDED_READERS: dict = {
        "ReadGridded": ReadGridded,
        "ReadMscwCtm": ReadMscwCtm,
        "ReadCAMS2_83": ReadCAMS2_83,
    }

    MODELS_WITH_KWARGS = [ReadMscwCtm]

    STATUS_CODES: dict[int, str] = {
        1: "SUCCESS",
        2: "NOT OK: Missing/invalid model variable",
        3: "NOT OK: Missing/invalid obs variable",
        4: "NOT OK: Failed to read model variable",
        5: "NOT OK: Colocation failed",
    }

    def __init__(self, colocation_setup: ColocationSetup | dict, **kwargs):
        if not colocation_setup:
            raise ValueError(
                "An instance ColocationSetup or a dict must be provided to Colocator."
            )
        if not isinstance(colocation_setup, ColocationSetup):
            colocation_setup = ColocationSetup(**colocation_setup)
            warnings.warn(
                DeprecationWarning(
                    "Future versions of Pyaerocom will require Colocator to ingest an instance of ColocationSetup."
                )
            )

        self.colocation_setup = colocation_setup
        self.logging: bool = True
        self._loaded_model_data: dict | None = {}
        self.data: dict = {}

        self._processing_status: list[str] = []
        self.files_written: list[str] = []

        self._model_reader: ReadGridded | ReadMscwCtm | ReadCAMS2_83 | None = None
        self._obs_reader: Any | None = None
        self._obs_is_vertical_profile: bool = False
        self.obs_filters: dict = colocation_setup.obs_filters.copy()

    @property
    def model_vars(self):
        """
        List of all model variables specified in config

        Note
        ----
        This method does not check if the variables are valid or available.

        Returns
        -------
        list
            list of all model variables specified in this setup.

        """
        ovars = self.colocation_setup.obs_vars
        model_vars = []
        for ovar in ovars:
            if ovar in self.colocation_setup.model_add_vars:
                model_vars.append(self.colocation_setup.model_add_vars[ovar])
            else:
                model_vars.append(ovar)

        for ovar, mvars in self.colocation_setup.model_add_vars.items():
            if ovar not in ovars:
                logger.warning(
                    f"Found entry in model_add_vars for obsvar {ovar} which "
                    f"is not specified in attr obs_vars, and will thus be "
                    f"ignored"
                )
            model_vars += mvars
        return model_vars

    @property
    def obs_is_ungridded(self):
        """
        bool: True if obs_id refers to an ungridded observation, else False
        """
        if self.colocation_setup.pyaro_config is not None:
            return True

        return True if self.colocation_setup.obs_id in get_all_supported_ids_ungridded() else False

    @property
    def obs_is_vertical_profile(self):
        """
        bool: True if obs_id refers to a VerticalProfile, else False
        """
        return self._obs_is_vertical_profile

    @obs_is_vertical_profile.setter
    def obs_is_vertical_profile(self, value):
        self._obs_is_vertical_profile = value

    @property
    def model_reader(self):
        """
        Model data reader
        """
        if self._model_reader is not None:
            if self._model_reader.data_id == self.colocation_setup.model_id:
                return self._model_reader
            logger.info(
                f"Reloading outdated model reader. ID of current reader: "
                f"{self._model_reader.data_id}. New ID: {self.colocation_setup.model_id}"
            )
        self._model_reader = self._instantiate_gridded_reader(what="model")
        self._loaded_model_data = {}
        return self._model_reader

    def _check_data_id_obs_reader(self):
        """
        Check if obs_reader is instantiated with correct obs ID.

        Returns
        -------
        bool
            True if current obs reader can be used for obs reading, else False.
        """
        reader = self._obs_reader
        if reader is None:
            return False
        elif self.obs_is_ungridded and self.colocation_setup.obs_id in reader.data_ids:
            return True
        elif self.colocation_setup.obs_id == reader.data_id:
            return True

    @property
    def obs_reader(self):
        """
        Observation data reader
        """

        if not self._check_data_id_obs_reader():
            if self.obs_is_ungridded:
                self._obs_reader = ReadUngridded(
                    data_ids=[self.colocation_setup.obs_id],
                    data_dirs=self.colocation_setup.obs_data_dir,
                    configs=[
                        self.colocation_setup.pyaro_config,
                    ],
                )
            else:
                self._obs_reader = self._instantiate_gridded_reader(what="obs")
        return self._obs_reader

    @property
    def output_dir(self):
        """
        str: Output directory for colocated data NetCDF files
        """
        loc = os.path.join(self.colocation_setup.basedir_coldata, self.get_model_name())
        if not os.path.exists(loc):
            logger.info(f"Creating dir {loc}")
            os.mkdir(loc)
        return loc

    @property
    def processing_status(self):
        head = ["Model Var", "Obs Var", "Status"]
        tab = []
        for mvar, ovar, key in self._processing_status:
            tab.append([mvar, ovar, self.STATUS_CODES[key]])
        return pd.DataFrame(tab, columns=head)

    def get_model_name(self):
        """
        Get name of model

        Note
        ----
        Not to be confused with :attr:`model_id` which is always the database
        ID of the model, while model_name can differ from that and is used for
        output files, etc.

        Raises
        ------
        AttributeError
            If neither model_id or model_name are set

        Returns
        -------
        str
            preferably :attr:`model_name`, else :attr:`model_id`

        """
        if self.colocation_setup.model_name is None:
            if self.colocation_setup.model_id is None:
                raise AttributeError(
                    "Neither model_name nor model_id are set. These must be set in ColocationSetup."
                )
            return self.colocation_setup.model_id
        return self.colocation_setup.model_name

    def get_obs_name(self):
        """
        Get name of obsdata source

        Note
        ----
        Not to be confused with :attr:`obs_id` which is always the database
        ID of the observation dataset, while obs_name can differ from that
        and is used for output files, etc.

        Raises
        ------
        AttributeError
            If neither obs_id or obs_name are set

        Returns
        -------
        str
            preferably :attr:`obs_name`, else :attr:`obs_id`

        """
        if self.colocation_setup.obs_name is None:
            if self.colocation_setup.obs_id is None:
                raise AttributeError("Neither obs_name nor obs_id are set")
            return self.colocation_setup.obs_id
        return self.colocation_setup.obs_name

    def get_model_data(self, model_var: str):
        if model_var in self._loaded_model_data:
            mdata = self._loaded_model_data[model_var]
            if mdata.data_id == self.colocation_setup.model_id:
                return mdata
        self._check_add_model_read_aux(model_var)
        mdata = self._read_gridded(var_name=model_var, is_model=True)
        self._loaded_model_data[model_var] = mdata
        return mdata

    def get_obs_data(self, obs_var: str):
        if self.obs_is_ungridded:
            return self._read_ungridded(obs_var)
        else:
            return self._read_gridded(obs_var, is_model=False)

    def get_start_str(self):
        if self.start is None:
            raise AttributeError("start time is not set")
        return to_datestring_YYYYMMDD(to_pandas_timestamp(self.start))

    def get_stop_str(self):
        if self.stop is None:
            raise AttributeError("stop time is not set")
        return to_datestring_YYYYMMDD(to_pandas_timestamp(self.stop))

    def prepare_run(self, var_list: list[str] | None = None) -> dict:
        """
        Prepare colocation run for current setup.

        Parameters
        ----------
        var_name : str, optional
            Variable name that is supposed to be analysed. The default is None,
            in which case all defined variables are attempted to be colocated.

        Raises
        ------
        AttributeError
            If no observation variables are defined (:attr:`obs_vars` empty).

        Returns
        -------
        vars_to_process : dict
            Mapping of variables to be processed, keys are model vars, values
            are obs vars.

        """

        if isinstance(self.colocation_setup.obs_vars, str):
            self.colocation_setup.obs_vars = (self.colocation_setup.obs_vars,)

        self._check_obs_vars_available()
        self._check_obs_filters()
        self._check_model_add_vars()
        self._check_set_start_stop()

        vars_to_process = self._find_var_matches()
        if var_list is not None:
            vars_to_process = self._filter_var_matches_varlist(vars_to_process, var_list)

        (vars_to_process, ts_types) = self._check_load_model_data(vars_to_process)

        if self.colocation_setup.save_coldata and not self.colocation_setup.reanalyse_existing:
            vars_to_process = self._filter_var_matches_files_not_exist(vars_to_process, ts_types)
        return vars_to_process

    def run(self, var_list: list[str] | None = None):
        """Perform colocation for current setup

        See also :func:`prepare_run`.

        Parameters
        ----------
        var_list : list, optional
            list of variables supposed to be analysed. The default is None,
            in which case all defined variables are attempted to be colocated.

        Returns
        -------
        dict
            nested dictionary, where keys are model variables, values are
            dictionaries comprising key / value pairs of obs variables and
            associated instances of :class:`ColocatedData`.
        """
        # MDA8 is a daily value so it doesn't make sense to calculate it if
        # no frequency is daily or coarser.
        calc_mda8 = False
        if self.colocation_setup.main_freq in ["daily", "monthly", "yearly"]:
            calc_mda8 = True
        if any(x in ["daily", "monthly", "yearly"] for x in self.colocation_setup.freqs):
            calc_mda8 = True

        data_out = defaultdict(lambda: dict())
        # TODO: see if the following could be solved via custom context manager
        try:
            vars_to_process = self.prepare_run(var_list)
        except Exception as ex:
            logger.exception(ex)
            if self.colocation_setup.raise_exceptions:
                self._print_processing_status()
                logger.critical(f"ABORTED: raise_exceptions is True: {traceback.format_exc()}\n")
                raise ex
            vars_to_process = {}
        self._print_coloc_info(vars_to_process)
        for mod_var, obs_var in vars_to_process.items():
            try:
                coldata = self._run_helper(
                    mod_var, obs_var
                )  # note this can be ColocatedData or ColocatedDataLists
                data_out[mod_var][obs_var] = coldata

                if calc_mda8 and (obs_var in MDA8_INPUT_VARS):
                    try:
                        mda8 = mda8_colocated_data(
                            coldata, obs_var=f"{obs_var}mda8", mod_var=f"{mod_var}mda8"
                        )
                    except ValueError as e:
                        logger.debug(e)
                    else:
                        self._save_coldata(mda8)
                        logger.info(
                            "Successfully calculated mda8 for [%s, %s].",
                            obs_var,
                            mod_var,
                        )
                        data_out[f"{mod_var}mda8"][f"{obs_var}mda8"] = mda8

                self._processing_status.append([mod_var, obs_var, 1])
            except Exception:
                msg = f"Failed to perform analysis: {traceback.format_exc()}\n"
                logger.warning(msg)
                self._processing_status.append([mod_var, obs_var, 5])
                if self.colocation_setup.raise_exceptions:
                    self._print_processing_status()
                    logger.critical("ABORTED: raise_exceptions is True\n")

                    raise ColocationError(traceback.format_exc())
        logger.info("Colocation finished")

        self._print_processing_status()
        if self.colocation_setup.keep_data:
            self.data = data_out
        return dict(data_out)

    def get_nc_files_in_coldatadir(self):
        """
        Get list of NetCDF files in colocated data directory

        Returns
        -------
        list
            list of NetCDF file paths found

        """
        mask = f"{self.output_dir}/*.nc"
        return glob.glob(mask)

    def get_available_coldata_files(self, var_list: list = None) -> list:
        self._check_set_start_stop()

        def check_meta_match(meta, **kwargs):
            for key, val in kwargs.items():
                if not meta[key] == val:
                    return False
            return True

        mname = self.get_model_name()
        oname = self.get_obs_name()
        model_vars = self.model_vars
        obs_vars = self.colocation_setup.obs_vars
        start, stop = self.get_start_str(), self.get_stop_str()
        valid = []
        all_files = self.get_nc_files_in_coldatadir()
        for file in all_files:
            try:
                meta = ColocatedData.get_meta_from_filename(file)
            except Exception:
                continue
            candidate = check_meta_match(
                meta, model_name=mname, obs_name=oname, start=start, stop=stop
            )
            if candidate and meta["model_var"] in model_vars and meta["obs_var"] in obs_vars:
                if var_list is None:
                    ok = True
                else:
                    ok = False
                    for var in var_list:
                        if meta["model_var"] == var or meta["obs_var"] == var:
                            ok = True
                            break
                if ok:
                    valid.append(file)

        return valid

    def _filter_var_matches_varlist(self, vars_to_process, var_list) -> dict:
        _vars_to_process = {}
        if isinstance(var_list, str):
            var_list = [var_list]
        for var_name in var_list:
            _subset = self._filter_var_matches_var_name(vars_to_process, var_name)
            _vars_to_process.update(**_subset)
        return _vars_to_process

    def _read_ungridded(self, var_name):
        """Helper to read UngriddedData

        Note
        ----
        reading is restricted to single variable UngriddedData objects here,
        due to multiple possibilities for variable specific obs filter
        definitions and because the colocation is done sequentially for each
        variable.

        Parameters
        ----------
        vars_to_read : str or list, optional
            variables that should be read from obs-network (:attr:`obs_id`)

        Returns
        -------
        UngriddedData
            loaded data object

        """
        obs_reader = self.obs_reader
        obs_filters_post = self._eval_obs_filters(var_name)

        obs_data = obs_reader.read(
            data_ids=[self.colocation_setup.obs_id],
            vars_to_retrieve=var_name,
            only_cached=self.colocation_setup.obs_cache_only,
            filter_post=obs_filters_post,
            **self.colocation_setup.read_opts_ungridded,
        )

        if self.colocation_setup.obs_remove_outliers:
            oor = self.colocation_setup.obs_outlier_ranges
            if var_name in oor:
                low, high = oor[var_name]
            else:
                var_info = const.VARS[var_name]
                low, high = var_info.minimum, var_info.maximum
            obs_data.remove_outliers(
                var_name, low=low, high=high, inplace=True, move_to_trash=False
            )
        return obs_data

    def _check_obs_filters(self):
        obs_vars = self.colocation_setup.obs_vars
        if any([x in self.colocation_setup.obs_filters for x in obs_vars]):
            # variable specific obs_filters
            for ovar in obs_vars:
                if ovar not in self.colocation_setup.obs_filters:
                    self.obs_filters[ovar] = {}

    def _check_load_model_data(self, var_matches):
        """
        Try to preload modeldata for input variable matches

        Note
        ----
        This will load all model fields for each obs variable in lazy mode, so
        should not require much storage. T

        Parameters
        ----------
        var_matches : dict
            dictionary specifying model / obs var pairs for colocation.

        Raises
        ------
        ColocationError
            If :attr:`raise_exceptions` is True and if one of the input
            model variables cannot be loaded.

        Returns
        -------
        filtered : dict
            `var_matches` filtered by entries for which model data could be
            successfully read.
        ts_types : dict
            data frequencies for each model variable that could be read. Those
            depend also on settings for :attr:`flex_ts_type`

        """
        filtered, ts_types = {}, {}
        for mvar, ovar in var_matches.items():
            try:
                mdata = self.get_model_data(mvar)
                filtered[mvar] = ovar
                ts_types[mvar] = mdata.ts_type
            except Exception as e:
                msg = f"Failed to load model data: {self.colocation_setup.model_id} ({mvar}). Reason {e}"
                logger.warning(msg)
                self._processing_status.append([mvar, ovar, 4])
                if self.colocation_setup.raise_exceptions:
                    raise ColocationError(msg)
        return filtered, ts_types

    def _filter_var_matches_files_not_exist(self, var_matches, ts_types):
        filtered = {}
        for mvar, ovar in var_matches.items():
            ts_type = ts_types[mvar]
            fname = self._coldata_savename(ovar, mvar, ts_type)
            fp = os.path.join(self.output_dir, fname)
            if not os.path.exists(fp):
                filtered[mvar] = ovar
        return filtered

    def _check_model_add_vars(self):
        for ovar, mvars in self.colocation_setup.model_add_vars.items():
            if not isinstance(mvars, list | tuple):
                raise ValueError("Values of model_add_vars need to be list or tuple")
            elif not all([isinstance(x, str) for x in mvars]):
                raise ValueError("Values of model_add_vars need to be list of strings")

    def _instantiate_gridded_reader(self, what):
        """
        Create reader for model or observational gridded data.

        Parameters
        ----------
        what : str
            Type of reader. ("model" or "obs")

        Returns
        -------
        Instance of reader class defined in self.SUPPORTED_GRIDDED_READERS
        """
        if what == "model":
            data_id = self.colocation_setup.model_id
            data_dir = self.colocation_setup.model_data_dir
        else:
            data_id = self.colocation_setup.obs_id
            data_dir = self.colocation_setup.obs_data_dir
        reader_class = self._get_gridded_reader_class(what=what)
        if what == "model" and reader_class in self.MODELS_WITH_KWARGS:
            reader = reader_class(
                data_id=data_id, data_dir=data_dir, **self.colocation_setup.model_kwargs
            )
        else:
            reader = reader_class(data_id=data_id, data_dir=data_dir)
        return reader

    def _get_gridded_reader_class(self, what):
        """Returns the class of the reader for gridded data."""
        try:
            reader = self.SUPPORTED_GRIDDED_READERS[self.colocation_setup.gridded_reader_id[what]]
        except KeyError as e:
            raise NotImplementedError(
                f"Reader {self.colocation_setup.gridded_reader_id[what]} is not supported: {e}"
            )
        return reader

    def _check_add_model_read_aux(self, model_var):
        if model_var not in self.colocation_setup.model_read_aux:
            return False
        info = self.colocation_setup.model_read_aux[model_var]
        if not isinstance(info, dict):
            raise ValueError(
                f"Invalid value for model_read_aux of variable {model_var}. "
                f"Need dictionary, got {info}"
            )
        elif not all([x in info for x in ["vars_required", "fun"]]):
            raise ValueError(
                f"Invalid value for model_read_aux dict of variable {model_var}. "
                f"Require keys vars_required and fun in dict, got {info}"
            )
        try:
            self.model_reader.add_aux_compute(var_name=model_var, **info)
        except DataCoverageError:
            return False
        return True

    def _check_obs_vars_available(self):
        if not len(self.colocation_setup.obs_vars) > 0:
            raise ColocationSetupError("no observation variables specified...")
        oreader = self.obs_reader
        if self.obs_is_ungridded:
            avail = oreader.get_vars_supported(
                self.colocation_setup.obs_id, self.colocation_setup.obs_vars
            )
        else:
            avail = []
            for ovar in self.colocation_setup.obs_vars:
                if oreader.has_var(ovar):
                    avail.append(ovar)

        if len(self.colocation_setup.obs_vars) > len(avail):
            for ovar in self.colocation_setup.obs_vars:
                if ovar not in avail:
                    logger.warning(
                        f"Obs variable {ovar} is not available in {self.colocation_setup.obs_id} "
                        f"and will be ignored"
                    )
                    self._processing_status.append([None, ovar, 3])

            if self.colocation_setup.raise_exceptions:
                invalid = [var for var in self.colocation_setup.obs_vars if var not in avail]
                invalid = "; ".join(invalid)
                raise DataCoverageError(
                    f"Invalid obs var(s) for {self.colocation_setup.obs_id}: {invalid}"
                )

            self.obs_vars = avail

    def _print_processing_status(self):
        mname = self.get_model_name()
        oname = self.get_obs_name()
        logger.info(f"Colocation processing status for {mname} vs. {oname}")
        logger.info(self.processing_status)

    def _filter_var_matches_var_name(self, var_matches, var_name):
        filtered = {}
        for mvar, ovar in var_matches.items():
            if mvar == var_name or ovar == var_name:
                filtered[mvar] = ovar
        if len(filtered) == 0:
            raise DataCoverageError(var_name)
        return filtered

    def _find_var_matches(self):
        """Find variable matches in model data for input obs variables"""

        # dictionary that will map model variables (keys) with observation
        # variables (values)
        var_matches = {}

        all_ok = True
        muv = self.colocation_setup.model_use_vars
        modreader = self.model_reader
        for ovar in self.colocation_setup.obs_vars:
            if ovar in muv:
                mvar = muv[ovar]
            else:
                mvar = ovar
            self._check_add_model_read_aux(mvar)
            if modreader.has_var(mvar):
                var_matches[mvar] = ovar
            else:
                self._processing_status.append([mvar, ovar, 2])
                all_ok = False

            if ovar in self.colocation_setup.model_add_vars:  # observation variable
                addvars = self.colocation_setup.model_add_vars[ovar]
                for addvar in addvars:
                    self._check_add_model_read_aux(addvar)
                    if modreader.has_var(addvar):
                        var_matches[addvar] = ovar
                    else:
                        self._processing_status.append([addvar, ovar, 2])
                        all_ok = False

        if not all_ok and self.colocation_setup.raise_exceptions:
            raise DataCoverageError("Some model variables are not available")

        return var_matches

    def _get_ts_type_read(self, var_name, is_model):
        """
        Get *desired* reading frequency for gridded reading

        Parameters
        ----------
        var_name : str
            Name of variable to be read.
        is_model : bool
            True if reading refers to model reading, else False (e.g. gridded
            satellite obs).

        Raises
        ------
        ValueError


        Returns
        -------
        str or None
            frequency to be read.

        """
        tst = self.colocation_setup.ts_type  # default
        if is_model and self.colocation_setup.model_ts_type_read is not None:
            tst = self.colocation_setup.model_ts_type_read
            if tst == "":
                tst = self.colocation_setup.ts_type
        elif not is_model and self.colocation_setup.obs_ts_type_read is not None:
            tst = self.obs_ts_type_read
        if isinstance(tst, dict):
            if var_name in tst:
                tst = tst[var_name]
            else:
                tst = self.colocation_setup.ts_type
        return tst

    def _read_gridded(self, var_name, is_model):
        start, stop = self.colocation_setup.start, self.colocation_setup.stop
        ts_type_read = self._get_ts_type_read(var_name, is_model)
        kwargs = {}
        if is_model:
            reader = self.model_reader
            vert_which = self.colocation_setup.obs_vert_type

            kwargs.update(**self.colocation_setup.model_kwargs)
            if self.colocation_setup.model_use_climatology:
                # overwrite start and stop to read climatology file for model
                start, stop = 9999, None
            if var_name in self.colocation_setup.model_read_opts:
                kwargs.update(self.colocation_setup.model_read_opts[var_name])
        else:
            reader = self.obs_reader
            vert_which = None
            ts_type_read = self.colocation_setup.obs_ts_type_read
            kwargs.update(self._eval_obs_filters(var_name))

        try:
            data = reader.read_var(
                var_name,
                start=start,
                stop=stop,
                ts_type=ts_type_read,
                vert_which=vert_which,
                flex_ts_type=self.colocation_setup.flex_ts_type,
                **kwargs,
            )
        except DataCoverageError:
            vert_which_alt = self._try_get_vert_which_alt(is_model, var_name)
            data = reader.read_var(
                var_name,
                start=start,
                stop=stop,
                ts_type=ts_type_read,
                flex_ts_type=self.colocation_setup.flex_ts_type,
                vert_which=vert_which_alt,
            )

        data = self._check_remove_outliers_gridded(data, var_name, is_model)
        return data

    def _try_get_vert_which_alt(self, is_model, var_name):
        if is_model:
            if self.colocation_setup.obs_vert_type in self.colocation_setup.OBS_VERT_TYPES_ALT:
                return self.colocation_setup.OBS_VERT_TYPES_ALT[
                    self.colocation_setup.obs_vert_type
                ]
        raise DataCoverageError(f"No alternative vert type found for {var_name}")

    def _check_remove_outliers_gridded(self, data: GriddedData, var_name: str, is_model: bool):
        if is_model:
            rm_outliers = self.colocation_setup.model_remove_outliers
            outlier_ranges = self.colocation_setup.model_outlier_ranges
        else:
            rm_outliers = self.colocation_setup.obs_remove_outliers
            outlier_ranges = self.colocation_setup.obs_outlier_ranges

        if len(outlier_ranges) > 0 and not rm_outliers:
            logger.warning(
                f"WARNING: Found definition of outlier ranges for {var_name} "
                f"({data.data_id}) but outlier removal is deactivated. Consider "
                f"checking your setup (note: model or obs outlier removal can be "
                f"activated via attrs. model_remove_outliers and remove_outliers, "
                f"respectively"
            )

        if rm_outliers:
            if var_name in outlier_ranges:
                low, high = outlier_ranges[var_name]
            else:
                var_info = const.VARS[var_name]
                low, high = var_info.minimum, var_info.maximum
            data.check_unit()
            data.remove_outliers(low, high, inplace=True)
        return data

    def _eval_obs_filters(self, var_name: str):
        obs_filters = self.obs_filters
        if var_name in obs_filters:
            # return obs_filters[var_name]
            obs_filters = obs_filters[var_name]
        if not isinstance(obs_filters, dict):
            raise AttributeError(
                f"Detected obs_filters attribute in Colocator class, "
                f"which is not a dictionary: {obs_filters}"
            )
        return obs_filters if len(obs_filters) > 0 else {}

    def _save_coldata(self, coldata: ColocatedData):
        """Helper for saving colocateddata"""
        obs_var, mod_var = coldata.metadata["var_name_input"]
        if mod_var in self.colocation_setup.model_rename_vars:
            mvar = self.colocation_setup.model_rename_vars[mod_var]
            logger.info(
                f"Renaming model variable from {mod_var} to {mvar} in "
                f"ColocatedData before saving to NetCDF."
            )
            coldata.rename_variable(mod_var, mvar, self.colocation_setup.model_id)
        else:
            mvar = mod_var

        if hasattr(coldata, "vertical_layer"):
            # save colocated vertical layer netCDF files with vertical layers in km
            if not Unit(coldata.data.altitude_units) == Unit("km"):
                start = Unit(coldata.data.altitude_units).convert(
                    coldata.vertical_layer["start"], other="km"
                )
                end = Unit(coldata.data.altitude_units).convert(
                    coldata.vertical_layer["end"], other="km"
                )
                vertical_layer = {"start": start, "end": end}
            else:
                vertical_layer = coldata.vertical_layer

            savename = self._coldata_savename(
                obs_var, mvar, coldata.ts_type, vertical_layer=vertical_layer
            )

        else:
            savename = self._coldata_savename(obs_var, mvar, coldata.ts_type)
        fp = coldata.to_netcdf(self.output_dir, savename=savename)
        self.files_written.append(fp)
        msg = f"WRITE: {fp}\n"
        logger.info(msg)

    def _eval_resample_how(self, model_var: str, obs_var: str):
        rshow = self.colocation_setup.resample_how
        if not isinstance(rshow, dict):
            return rshow

        if obs_var in rshow:
            return rshow[obs_var]
        elif model_var in rshow:
            return rshow[model_var]
        else:
            return None

    def _infer_start_stop_yr_from_model_reader(self):
        """
        Infer start / stop year for colocation from gridded model reader

        Sets :attr:`start` and :attr:`stop`

        """
        # get sorted list of years available in model data (files with year
        # 9999 denote climatological data)
        yrs_avail = self.model_reader.years_avail
        if self.colocation_setup.model_use_climatology:
            if 9999 not in yrs_avail:
                raise DataCoverageError("No climatology files available")
            first, last = 9999, None
        else:
            if 9999 in yrs_avail:
                yrs_avail = [x for x in yrs_avail if not x == 9999]
            first, last = yrs_avail[0], yrs_avail[-1]
            if first == last:
                last = None
        self.start = first
        self.stop = last

    def _check_set_start_stop(self):
        if self.colocation_setup.start is None:
            self._infer_start_stop_yr_from_model_reader()
            start, stop = self.start, self.stop
        else:
            start, stop = self.colocation_setup.start, self.colocation_setup.stop
        if self.colocation_setup.model_use_climatology:
            if self.colocation_setup.stop is not None or not isinstance(
                self.colocation_setup.start, int
            ):
                raise ColocationSetupError(
                    "Conflict: only single year analyses are support for model "
                    'climatology fields, please specify "start" as integer '
                    'denoting the year, and set "stop"=None'
                )
        self.start, self.stop = start_stop(start, stop)

    def _coldata_savename(self, obs_var, mod_var, ts_type, **kwargs):
        """Get filename of colocated data file for saving"""
        if "vertical_layer" in kwargs:
            vertical_layer = kwargs["vertical_layer"]
        else:
            vertical_layer = None
        name = ColocatedData._aerocom_savename(
            obs_var=obs_var,
            obs_id=self.get_obs_name(),
            mod_var=mod_var,
            mod_id=self.get_model_name(),
            start_str=self.get_start_str(),
            stop_str=self.get_stop_str(),
            ts_type=ts_type,
            filter_name=self.colocation_setup.filter_name,
            vertical_layer=vertical_layer,
        )
        return f"{name}.nc"

    def _get_colocation_ts_type(self, model_ts_type, obs_ts_type=None):
        chk = [self.colocation_setup.ts_type, model_ts_type]
        if obs_ts_type is not None:
            chk.append(obs_ts_type)
        return get_lowest_resolution(*chk)

    @property
    def _colocation_func(self):
        """
        Function used for colocation

        Returns
        -------
        callable
            function the performs co-location operation

        """

        if self.obs_is_vertical_profile:
            return colocate_vertical_profile_gridded
        if self.obs_is_ungridded:
            return colocate_gridded_ungridded
        else:
            return colocate_gridded_gridded

    def _prepare_colocation_args(self, model_var: str, obs_var: str):
        model_data = self.get_model_data(model_var)
        obs_data = self.get_obs_data(obs_var)

        if self.colocation_setup.harmonise_units:
            model_data, obs_data = harmonise_units(
                model_data, obs_data, var=model_var, var_ref=obs_var, inplace=True
            )

        if getattr(obs_data, "is_vertical_profile", None):
            self.obs_is_vertical_profile = obs_data.is_vertical_profile

        rshow = self._eval_resample_how(model_var, obs_var)

        if self.colocation_setup.model_use_climatology:
            baseyr = self.start.year
        else:
            baseyr = None
        # input args shared between all colocation functions
        args = dict(
            data=model_data,
            data_ref=obs_data,
            start=self.start,
            stop=self.stop,
            filter_name=self.colocation_setup.filter_name,
            regrid_res_deg=self.colocation_setup.regrid_res_deg,
            update_baseyear_gridded=baseyr,
            min_num_obs=self.colocation_setup.min_num_obs,
            colocate_time=self.colocation_setup.colocate_time,
            resample_how=rshow,
            add_meta_keys=[],
        )
        # check if the station_type key has been passed to the ungridded data object
        # (not all readers may do that, currently only the CAMS2_83 reader does)
        if isinstance(obs_data, UngriddedDataContainer):
            add_meta_keys = []
            are_there_station_types = {"station_type" in d for d in obs_data.metadata.values()}
            if are_there_station_types == {True, False}:
                raise ValueError("some stations have `station_type` metadata while others do not")
            if are_there_station_types == {True}:  # all have station_type
                add_meta_keys.append("station_type")

            if any("display_name" in d for d in obs_data.metadata.values()):
                add_meta_keys.append("display_name")
            args["add_meta_keys"] = add_meta_keys

        if self.obs_is_ungridded:
            ts_type = self._get_colocation_ts_type(model_data.ts_type)
            args.update(
                ts_type=ts_type,
                var_ref=obs_var,
                use_climatology_ref=self.colocation_setup.obs_use_climatology,
            )
        else:
            ts_type = self._get_colocation_ts_type(model_data.ts_type, obs_data.ts_type)
            args.update(ts_type=ts_type)
        if self.obs_is_vertical_profile:
            args.update(
                colocation_layer_limits=self.colocation_setup.colocation_layer_limits,
                profile_layer_limits=self.colocation_setup.profile_layer_limits,
            )
        return args

    def _check_dimensionality(self, args):
        mdata = args["data"]
        odata = args["data_ref"]
        from pyaerocom.exceptions import DataDimensionError
        from pyaerocom.griddeddata import GriddedData

        if mdata.ndim == 4 and self.colocation_setup.obs_vert_type == "Surface":
            mdata = mdata.extract_surface_level()
            args["data"] = mdata

        if isinstance(odata, GriddedData):
            if odata.ndim == 4 and self.colocation_setup.obs_vert_type == "Surface":
                odata = odata.extract_surface_level()
                args["data_ref"] = odata
            elif odata.ndim > 3:
                raise DataDimensionError(
                    f"cannot co-locate model data with more than 3 dimensions: {odata}"
                )
        return args

    def _run_helper(self, model_var: str, obs_var: str):
        logger.info(
            f"Running {self.colocation_setup.model_id} ({model_var}) vs. {self.colocation_setup.obs_id} ({obs_var})"
        )
        args = self._prepare_colocation_args(model_var, obs_var)
        args = self._check_dimensionality(args)
        coldata = self._colocation_func(**args)

        if isinstance(coldata, ColocatedData):
            coldata.data.attrs["model_name"] = self.get_model_name()
            coldata.data.attrs["obs_name"] = self.get_obs_name()
            coldata.data.attrs["vert_code"] = self.colocation_setup.obs_vert_type

            coldata.data.attrs.update(**self.colocation_setup.add_meta)

            if self.colocation_setup.zeros_to_nan:
                coldata = coldata.set_zeros_nan()
            if self.colocation_setup.model_to_stp:
                coldata = correct_model_stp_coldata(coldata)
            if self.colocation_setup.save_coldata:
                self._save_coldata(coldata)

        elif isinstance(coldata, ColocatedDataLists):  # look into intertools chain.from_iterable
            for i_list in coldata:
                for coldata_obj in i_list:
                    coldata_obj.data.attrs["model_name"] = self.get_model_name()
                    coldata_obj.data.attrs["obs_name"] = self.get_obs_name()
                    coldata_obj.data.attrs["vert_code"] = self.colocation_setup.obs_vert_type
                    coldata_obj.data.attrs.update(**self.colocation_setup.add_meta)
                    if self.colocation_setup.zeros_to_nan:
                        coldata_obj = coldata_obj.set_zeros_nan()
                    if self.colocation_setup.model_to_stp:  # TODO: check is this needs modifying
                        coldata = correct_model_stp_coldata(coldata_obj)
                    if self.colocation_setup.save_coldata:
                        self._save_coldata(coldata_obj)

        else:
            raise Exception(
                f"Invalid coldata type returned by colocation function {self._colocation_func}"
            )

        return coldata

    def _print_coloc_info(self, var_matches):
        if not var_matches:
            logger.info("Nothing to colocate")
            return
        key_vals = []
        for key, val in var_matches.items():
            key_vals.append(f"{key}\t{val}")
        logger.info(
            "The following variable combinations will be colocated\nMODEL-VAR\tOBS-VAR\n"
            + "\n".join(key_vals)
        )



from __future__ import annotations

from datetime import datetime

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from .data_access import TEST_DATA

EMEP_DATA_PATH = TEST_DATA["MODELS"].path / "EMEP_2017"


@pytest.fixture(scope="session")
def path_emep() -> dict[str, str]:
    """dictionary containing EMEP test data"""
    paths = dict(
        daily=EMEP_DATA_PATH / "Base_day.nc",
        monthly=EMEP_DATA_PATH / "Base_month.nc",
        yearly=EMEP_DATA_PATH / "Base_fullrun.nc",
        data_dir=EMEP_DATA_PATH,
    )

    return {key: str(path) for key, path in paths.items()}


def create_fake_MSCWCtm_data(year="2019", numval=1, tst=None):
    if tst is None:
        tst = "monthly"
    from pyaerocom import TsType

    _lats_fake = np.linspace(30, 82, 10)
    _lons_fake = np.linspace(-25, 90, 15)

    start = datetime.strptime(f"{year}-01-01", "%Y-%m-%d")
    stop = datetime.strptime(f"{year}-12-31", "%Y-%m-%d")
    _time_fake = pd.date_range(start, stop, freq=TsType(tst).to_pandas_freq())
    sh = (len(_time_fake), len(_lats_fake), len(_lons_fake))
    _data_fake = numval * np.ones(sh)

    coords = {"time": _time_fake, "latitude": _lats_fake, "longitude": _lons_fake}

    dims = ["time", "latitude", "longitude"]

    arr = xr.DataArray(data=_data_fake, coords=coords, dims=dims)
    arr.time.encoding["units"] = "hours since 1900-01-01 00:00:00"
    arr.time.encoding["calendar"] = "gregorian"
    arr.time.encoding["standard_name"] = "time"

    return arr


@pytest.fixture
def fake_aod_MSCWCtm_data_monthly_2010(tmp_path) -> str:
    path = tmp_path / "EMEP_fake" / "2010"

    if not path.exists():
        path.mkdir(parents=True)
    data = create_fake_MSCWCtm_data(year=2010, numval=1)

    var_name = "AOD_550nm"
    units = "1"
    ds = xr.Dataset()

    ds[var_name] = data
    ds[var_name].attrs.update(units=units, var_name=var_name)

    ds.to_netcdf(path=path / "Base_month.nc")

    return str(path)



"""
High level test methods that check AasEtAl time series data for selected stations
"""

from __future__ import annotations

from pathlib import Path

import pandas as pd
import pytest
from numpy.testing import assert_almost_equal

from pyaerocom import const
from pyaerocom.io.read_aasetal import ReadAasEtal
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.units.units_helpers import convert_unit
from tests.conftest import lustre_unavail

VARUNITS = {
    "concso2": "ug m-3",
    "concso4": "ug m-3",
    "wetso4": "kg m-2 s-1",
    "concso4pr": "g m-3",
    "pr": "mm",
}

DATA_ID = "GAWTADsubsetAasEtAl"

FILENAMES = ["monthly_so2.csv", "monthly_so4_aero.csv", "monthly_so4_precip.csv"]

VARS = ["concso2", "concso4", "concso4pr", "pr", "wetso4"]


@pytest.fixture
def data_path() -> Path:
    data_dir = const.OBSLOCS_UNGRIDDED[const.GAWTADSUBSETAASETAL_NAME]
    path = Path(data_dir)
    assert path.exists()
    assert path.is_dir()
    return path


@pytest.fixture
def data_paths(data_path: Path) -> list[Path]:
    paths = [data_path / file for file in FILENAMES]
    assert all(path.exists() for path in paths)
    return paths


@lustre_unavail
def test__get_time_stamps(data_paths: list[Path]):
    df = pd.read_csv(data_paths[0], sep=",", low_memory=False)
    reader = ReadAasEtal()
    timestamps = reader._get_time_stamps(df[:10])

    assert str(timestamps[0]) == "1997-09-01T00:00:00"
    assert str(timestamps[-1]) == "1998-09-01T00:00:00"


@lustre_unavail
def test_reader(data_path: Path):
    reader = ReadAasEtal(DATA_ID)
    assert reader.data_id == DATA_ID
    assert reader.data_dir == str(data_path)
    assert reader.PROVIDES_VARIABLES == ["concso2", "concso4", "pr", "wetso4", "concso4pr"]
    filenames = [Path(file).name for file in reader.get_file_list()]
    assert filenames == FILENAMES


@pytest.fixture(scope="session")
def aasetal_data() -> UngriddedData:
    """read expensive dataset"""
    reader = ReadAasEtal()
    return reader.read()


@lustre_unavail
def test_aasetal_data(aasetal_data: UngriddedData):
    data = aasetal_data
    assert len(data.station_name) == 890
    assert len(data.unique_station_names) == 667
    assert data.shape == (416243, 12)
    assert data.contains_vars == VARS
    assert data.contains_instruments == [
        "3_stage_filterpack",
        "passive_sampler",
        "abs_solution",
        "monitor",
        "filter_1pack",
        "filterpack",
        "2_stage_filterpack",
        "filter_denuder_sampler",
        "IMPROVE_PM2.5",
        "filter-1pack",
        "filter_3pack",
        "filter_2pack",
        "pm10_sampler",
        "filter-3pack",
        "wet only",
        "bulk",
        "bulk ",
        "wet-only",
    ]


@lustre_unavail
def test_aasetal_data_correct_units(aasetal_data: UngriddedData):
    tested = []
    stats = []
    for meta_key, meta in aasetal_data.metadata.items():
        for var, info in meta["var_info"].items():
            if var in tested:
                # test each variable only once
                continue
            assert info["units"] == VARUNITS[var]
            tested.append(var)
            stats.append(meta["station_name"])
        if len(tested) == len(VARS):
            break

    assert meta_key == 520
    assert stats == [
        "Abington",
        "Abington",
        "Abington (CT15)",
        "Abington (CT15)",
        "Abington (CT15)",
    ]


# TODO: test wetso4 (needs proper unit conversion)
testdata = [
    (0, "Yellowstone NP", "concentration_ugS/m3", "concso2"),
    (1, "Payerne", "concentration_ugS/m3", "concso4"),
    # (2, 'Abington (CT15)', 'deposition_kgS/ha', 'wetso4')
]


@lustre_unavail
@pytest.mark.parametrize("filenum,station_name,colname,var_name", testdata)
def test_reading_routines(
    aasetal_data: UngriddedData, data_paths: list[Path], filenum, station_name, colname, var_name
):
    UNITCONVERSION = ReadAasEtal().UNITCONVERSION

    df = pd.read_csv(data_paths[filenum], sep=",", low_memory=False)
    subset = df[df.station_name == station_name]
    # values in original units
    vals = subset[colname].astype(float).values
    from_unit, to_unit = UNITCONVERSION[var_name]
    should_be = convert_unit(vals, from_unit=from_unit, to_unit=to_unit, var_name=var_name).mean()

    actual = aasetal_data.to_station_data(station_name, var_name)[var_name].values.mean()

    if var_name == "wetso4":
        raise NotImplementedError
        # from pyaerocom.helpers import get_tot_number_of_seconds

        # numsecs = get_tot_number_of_seconds(ts_type='monthly',
        #                                    dtime=station_group['dtime'])
        # stat[var] = stat[var]/numsecs
        # to_unit = 'kg m-2 s-1'
    assert_almost_equal(should_be, actual)



from pydantic import BaseModel
import datetime
from getpass import getuser

from pyaerocom import __version__


class ExperimentInfo(BaseModel):
    exp_id: str
    exp_name: str = ""
    exp_descr: str = ""
    public: bool = False
    exp_pi: str = getuser()
    pyaerocom_version: str = __version__
    creation_date: str = f"{datetime.datetime.now(datetime.timezone.utc):%Y-%m-%dT%H:%M:%S.%fZ}"



from pydantic import BaseModel


class UnitsSetup(BaseModel):
    """
    Unit setup options

    Attributes
    ----------
    units : dict[str, str]
        A dictionary specifying a mapping from variable name (eg. concno2) to unit. All unit conversion
        by pyaerocom will use this unit for harmonisation if provided. If not provided, the unit specified
        in variables.ini will be used.
    """

    units: dict[str, str] = {}



#!/usr/bin/env python3
"""
Goal
"""

import os
import shutil
from functools import reduce
from pathlib import Path

import numpy as np

import pyaerocom as pya

OUTBASE = Path(pya.const._TESTDATADIR).joinpath("obsdata")

if not OUTBASE.exists():
    OUTBASE.mkdir()

MIN_NUM_VALID = 300

IDS = {
    "sun": "AeronetSunV3Lev2.daily",
    "sda": "AeronetSDAV3Lev2.daily",
    "inv": "AeronetInvV3Lev2.daily",
}

NETWORKS = {"sun": "od550aer", "sda": "od550lt1aer", "inv": "abs550aer"}

filters = [
    ("altitude", [3000, 10000], 2),
    ("altitude", [0, 1000], 10),
    ("region_id", "OCN", 2),
    ("station_name", "T*", 10),
]

revision_files = {}
if __name__ == "__main__":
    loaded = {}
    for name, varlist in NETWORKS.items():
        reader = pya.io.ReadUngridded()
        _data = reader.read(IDS[name], varlist)
        loaded[name] = _data
        r = reader.get_reader()
        revision_files[name] = Path(r.DATASET_PATH).joinpath(r.REVISION_FILE)

    use_stats = []

    for attr, val, maxnum in filters:
        subsets = {}
        statnames = []

        for name, data in loaded.items():
            subset = data.apply_filters(**{attr: val})

            subsets[name] = subset
            statnames.append(subset.unique_station_names)

        # same_stats = np.intersect1d(*statnames)

        same_stats = reduce(np.intersect1d, statnames)
        print(same_stats)

        stats_ok = []
        for statname in same_stats:
            if len(stats_ok) == maxnum:
                print("YEAH")
                break
            elif statname in use_stats:
                continue
            statok = True
            for name, subset in subsets.items():
                var = NETWORKS[name]
                try:
                    stat = subset.to_station_data(statname)
                except pya.exceptions.DataCoverageError:
                    break
                data = stat[var]
                numvalid = np.sum(~np.isnan(data))
                if not numvalid > MIN_NUM_VALID:
                    statok = False
                    break
            if statok:
                stats_ok.append(statname)
        if len(stats_ok) == 0:
            raise Exception
        print(f"Found {len(stats_ok)} common sites for filter {attr}: {val}")

        use_stats.extend(stats_ok)

    files = {"sun": [], "sda": [], "inv": []}

    for name, data in loaded.items():
        data_id = IDS[name]
        outdir = OUTBASE.joinpath(data_id)
        # make sure to remove old data
        if outdir.exists():
            print(f"REMOVING EXISTING DATA FOR {data_id}")
            shutil.rmtree(outdir)
        outdir.mkdir()

        renamed = outdir.joinpath("renamed")
        renamed.mkdir()

        for statname in use_stats:
            idx = data._find_station_indices(statname)

            if len(idx) != 1:
                raise Exception
            fname = Path(data.metadata[idx[0]]["filename"])
            if not fname.exists():
                raise Exception
            dest = renamed.joinpath(fname.name)
            shutil.copy(str(fname), dest)
            files[name].append(str(fname))

        shutil.copy(revision_files[name], outdir.joinpath("Revision.txt"))

    for name, filelist in files.items():
        print(f"Copied {len(filelist)} files for {name} into {os.path.dirname(filelist[0])}")



import numpy as np
import pytest
import xarray as xr

from pyaerocom.aeroval import EvalSetup  # , ExperimentProcessor
from pyaerocom.aeroval.fairmode_statistics import SPECIES
from pyaerocom.scripts.cams2_83.engine import CAMS2_83_Engine
from tests.fixtures.collocated_data import COLDATA


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test__calc_forecast_target_MQI_vectorized(eval_config: dict):
    example_coldata = COLDATA["tm5_aeronet"]()

    var_name = "concno2"

    # rename variable to pretend it's a fairmode species
    example_coldata.data = example_coldata.data.assign_attrs(var_name=[var_name, var_name])

    # create dummy persistence model coldata
    example_persistence_coldata = example_coldata

    start = example_persistence_coldata.data["time"].values[0] - np.timedelta64(1, "D")
    end = example_persistence_coldata.data["time"].values[-1]
    example_persistence_coldata.data = example_persistence_coldata.data.reindex(
        {"time": xr.date_range(start, end, freq="h")}
    )

    # set dummy values so that mod_vals will be all 2, obs_vals and p_mod_vals all 1
    # this means that mqi = 1/rmse_persistence and the expected value is just 1.0 / uncertainty_p_obs_val
    p_mod_val = obs_val = 1
    mod_val = 2
    factor = SPECIES[var_name]["alpha"] ** 2 * SPECIES[var_name]["RV"] ** 2
    uncertainty_p_obs_val = SPECIES[var_name]["UrRV"] * np.sqrt(
        (1 - SPECIES[var_name]["alpha"] ** 2) * p_mod_val**2 + factor
    )
    expected_val = 1.0 / uncertainty_p_obs_val

    example_coldata.data[1] = example_coldata.data[1].where(False, mod_val)
    example_coldata.data[0] = example_coldata.data[0].where(False, obs_val)
    example_persistence_coldata.data[0] = example_persistence_coldata.data[0].where(
        False, p_mod_val
    )

    setup = EvalSetup(**eval_config)
    c23engine = CAMS2_83_Engine(setup)

    mqi_vectorized = c23engine._calc_forecast_target_MQI_vectorized(
        example_coldata,
        example_persistence_coldata,
        "concno2",
        1,
        min_num_obs=eval_config["min_num_obs"],
    )

    stations_list = example_coldata.coords["station_name"].values.tolist()
    assert len(mqi_vectorized) == len(stations_list)
    for station in stations_list:
        assert len(mqi_vectorized[station]) == 3
        assert mqi_vectorized[station][0].dtype == mqi_vectorized[station][1].dtype == "float64"
        assert (
            np.round(1.0 / mqi_vectorized[station][0], 8)
            == np.round(mqi_vectorized[station][1], 8)
            == np.round(expected_val, 8)
        )



# isort:skip_file
import logging

logger = logging.getLogger(__name__)

# Submodule
from . import cams2_83

from .aerocom_browser import AerocomBrowser
from .ebas_file_index import EbasFileIndex, EbasSQLRequest

# low level EBAS I/O routines
from .ebas_nasa_ames import EbasNasaAmesFile
from .file_conventions import FileConventionRead
from .read_aasetal import ReadAasEtal

# read base classes
from .readgridded import ReadGridded
from .readungriddedbase import ReadUngriddedBase
from .readungridded import ReadUngridded

# Pyaerocom reading interface classes
from .read_aeronet_invv3 import ReadAeronetInvV3
from .read_aeronet_sdav3 import ReadAeronetSdaV3
from .read_aeronet_sunv3 import ReadAeronetSunV3
from .read_airnow import ReadAirNow
from .read_earlinet import ReadEarlinet
from .read_ebas import ReadEbas
from .read_eea_aqerep import ReadEEAAQEREP
from .read_eea_aqerep_v2 import ReadEEAAQEREP_V2
from .read_eprofile import ReadEprofile
from pyaerocom.io.cams2_83.reader import ReadCAMS2_83

# Pyaro classes
from .pyaro.read_pyaro import ReadPyaro
from .pyaro.pyaro_config import PyaroConfig

from . import netcdf_fix
